# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for scaling noise by inserting identity layers."""

import pytest
from cirq import Circuit, LineQubit, ops

from mitiq.utils import _equal
from mitiq.zne.scaling.identity_insertion import (
    UnscalableCircuitError,
    _calculate_id_layers,
    insert_id_layers,
)


@pytest.mark.parametrize("scale_factor", (1, 2, 3, 4, 5, 6))
def test_id_layers_whole_scale_factor(scale_factor):
    """Tests if n-1 identity layers are inserted uniformly when
    the intended scale factor is n."""
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.X.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    scaled_circ = insert_id_layers(circ, scale_factor=scale_factor)
    num_layers = scale_factor - 1
    expected_circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.I.on_each(*qreg)] * num_layers,
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.X.on(qreg[2])],
        [ops.I.on_each(*qreg)] * num_layers,
        [ops.TOFFOLI.on(*qreg)],
        [ops.I.on_each(*qreg)] * num_layers,
    )
    assert _equal(scaled_circ, expected_circ)


def test_scale_with_intermediate_measurements_raises_error():
    """Tests scaling function raise an error on circuits with
    intermediate measurements.
    """
    qbit = LineQubit(0)
    circ = Circuit([ops.H.on(qbit)], [ops.measure(qbit)], [ops.T.on(qbit)])
    with pytest.raises(
        UnscalableCircuitError,
        match="Circuit contains intermediate measurements",
    ):
        insert_id_layers(circ, scale_factor=3.0)


def test_scaling_with_terminal_measurement():
    """Checks if the circuit with a terminal measurement is
    scaled with identity layers as expected.
    """
    qbit = LineQubit(0)
    input_circ = Circuit(
        [ops.H.on(qbit)], [ops.T.on(qbit)], [ops.measure(qbit)]
    )
    scaled_circ = insert_id_layers(input_circ, scale_factor=3.0)
    expected_circ = Circuit(
        [ops.H.on(qbit)],
        [ops.I.on(qbit)] * 2,
        [ops.T.on(qbit)],
        [ops.I.on(qbit)] * 2,
        [ops.measure(qbit)],
    )
    assert _equal(scaled_circ, expected_circ)


def test_calculate_id_layers_diff_scale_factor():
    """Checks if the partial layers to be inserted are 0 for a full
    scale factor and may or may not be 0 otherwise.

    Also checks if an error is raised for an invalid scale factor
    value.
    """
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.X.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    circ_depth = len(circ)
    full_scale_factor = 3
    id_layers_full_scale = _calculate_id_layers(circ_depth, full_scale_factor)
    num_partial_layers_full_scale_factor = id_layers_full_scale[-1]
    assert num_partial_layers_full_scale_factor == 0

    float_scale_factor_list = [1.3, 2.6, 3.77, 4.8, 5.9]
    for i in float_scale_factor_list:
        float_scale_factor = i
        id_layers_float_scale = _calculate_id_layers(
            circ_depth, float_scale_factor
        )

        num_partial_layers_float_scale_factor = id_layers_float_scale[-1]
        assert num_partial_layers_float_scale_factor >= 0

    bad_scale_factor_list = [-1.3, 0, -3, 0.76]
    for i in bad_scale_factor_list:
        with pytest.raises(ValueError, match="Requires scale_factor >= 1"):
            _calculate_id_layers(circ_depth, i)
            insert_id_layers(circ, i)


@pytest.mark.parametrize(
    "intended_scale_factor",
    (1, 1.1, 1.3, 1.7, 1.9, 2, 3.1, 3.6, 3.9, 3, 4, 5, 6),
)
def test_compare_scale_factor(intended_scale_factor):
    """tests if the intended scale factor is approximately close to the
    actual scale factor.
    """
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(*qreg)],
        [ops.CNOT.on(qreg[0], qreg[1])],
        [ops.X.on(qreg[2])],
        [ops.TOFFOLI.on(*qreg)],
    )
    scaled = insert_id_layers(circ, intended_scale_factor)
    achieved_scale_factor = len(scaled) / len(circ)
    assert achieved_scale_factor <= intended_scale_factor
    assert achieved_scale_factor >= intended_scale_factor - 1



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for the data regression portion of Clifford data regression."""

import numpy as np
import pytest

from mitiq.cdr.data_regression import linear_fit_function


@pytest.mark.parametrize("nvariables", [1, 3])
def test_linear_fit_function(nvariables):
    a = np.random.RandomState(1).rand(nvariables)
    b = np.ones(nvariables + 1)
    assert linear_fit_function(a, b) == sum(a) + 1



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for generating (near) Clifford circuits."""

import cirq
import numpy as np
import pytest
from cirq.circuits import Circuit

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.cdr._testing import random_x_z_cnot_circuit
from mitiq.cdr.clifford_training_data import (
    _map_to_near_clifford,
    _replace,
    _select,
    generate_training_circuits,
)
from mitiq.cdr.clifford_utils import count_non_cliffords, is_clifford
from mitiq.interface import convert_from_mitiq


def test_generate_training_circuits():
    circuit = random_x_z_cnot_circuit(
        cirq.LineQubit.range(3), n_moments=5, random_state=1
    )
    assert not is_clifford(circuit)

    (clifford_circuit,) = generate_training_circuits(
        circuit, num_training_circuits=1, fraction_non_clifford=0.0
    )
    assert is_clifford(clifford_circuit)


@pytest.mark.parametrize("circuit_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_generate_training_circuits_any_qprogram(circuit_type):
    circuit = random_x_z_cnot_circuit(
        cirq.LineQubit.range(3), n_moments=5, random_state=1
    )
    circuit = convert_from_mitiq(circuit, circuit_type)

    (clifford_circuit,) = generate_training_circuits(
        circuit, num_training_circuits=1, fraction_non_clifford=0.0
    )
    assert is_clifford(clifford_circuit)


@pytest.mark.parametrize("method", ("uniform", "gaussian"))
def test_select_all(method):
    q = cirq.LineQubit(0)
    ops = [cirq.ops.rz(0.01).on(q), cirq.ops.rz(-0.77).on(q)]
    indices = _select(
        ops, 0.0, method=method, random_state=np.random.RandomState(1)
    )
    assert np.allclose(indices, np.array(list(range(len(ops)))))


@pytest.mark.parametrize("method", ("uniform", "gaussian"))
def test_select_some(method):
    n = 10  # Number to select.
    q = cirq.GridQubit(1, 1)
    ops = [cirq.ops.rz(a).on(q) for a in np.random.randn(n)]
    indices = _select(ops, fraction_non_clifford=0.5, method=method)
    assert len(indices) == n // 2


def test_select_bad_method():
    with pytest.raises(ValueError, match="Arg `method_select` must be"):
        _select([], fraction_non_clifford=0.0, method="unknown method")


@pytest.mark.parametrize("method", ("closest", "uniform", "gaussian"))
def test_replace(method):
    q = cirq.LineQubit(0)
    ops = [cirq.ops.rz(0.01).on(q), cirq.ops.rz(-0.77).on(q)]

    new_ops = _replace(
        non_clifford_ops=ops,
        method=method,
        random_state=np.random.RandomState(1),
    )

    assert len(new_ops) == len(ops)
    assert all(cirq.has_stabilizer_effect(op) for op in new_ops)


def test_map_to_near_clifford():
    q = cirq.LineQubit(0)
    ops = [cirq.ops.rz(np.pi / 2.0 + 0.01).on(q), cirq.ops.rz(-0.22).on(q)]

    new_ops = _map_to_near_clifford(
        ops,
        fraction_non_clifford=0.0,
        method_select="uniform",
        method_replace="uniform",
        random_state=np.random.RandomState(2),
    )
    expected_ops = [cirq.rz(np.pi * 1).on(q), cirq.rz(np.pi * 1.5).on(q)]
    assert new_ops == expected_ops


def test_generate_training_circuits_bad_methods():
    with pytest.raises(ValueError):
        generate_training_circuits(
            Circuit(cirq.ops.rx(0.5).on(cirq.LineQubit(0))),
            num_training_circuits=1,
            fraction_non_clifford=0.0,
            method_select="unknown select method",
        )

    with pytest.raises(ValueError):
        generate_training_circuits(
            Circuit(cirq.ops.rx(0.5).on(cirq.LineQubit(0))),
            num_training_circuits=1,
            fraction_non_clifford=0.0,
            method_replace="unknown replace method",
        )


def test_generate_training_circuits_with_clifford_circuit():
    circuit = Circuit(cirq.ops.rx(0.0).on(cirq.LineQubit(0)))
    assert generate_training_circuits(
        circuit,
        num_training_circuits=2,
        fraction_non_clifford=0.0,
    ) == [circuit, circuit]


@pytest.mark.parametrize("method_select", ["uniform", "gaussian"])
@pytest.mark.parametrize("method_replace", ["uniform", "gaussian", "closest"])
@pytest.mark.parametrize(
    "kwargs", [{}, {"sigma_select": 0.5, "sigma_replace": 0.5}]
)
def test_generate_training_circuits_mega(
    method_select, method_replace, kwargs
):
    circuit = random_x_z_cnot_circuit(qubits=4, n_moments=10, random_state=1)
    num_train = 10
    fraction_non_clifford = 0.1

    train_circuits = generate_training_circuits(
        circuit,
        num_training_circuits=num_train,
        fraction_non_clifford=0.1,
        random_state=np.random.RandomState(13),
        method_select=method_select,
        method_replace=method_replace,
        **kwargs,
    )
    assert len(train_circuits) == num_train

    for train_circuit in train_circuits:
        assert set(train_circuit.all_qubits()) == set(circuit.all_qubits())
        assert count_non_cliffords(train_circuit) == int(
            round(fraction_non_clifford * count_non_cliffords(circuit))
        )


@pytest.mark.parametrize("method", ["uniform", "gaussian"])
def test_select(method):
    q = cirq.NamedQubit("q")
    non_clifford_ops = [
        cirq.ops.rz(a).on(q) for a in np.linspace(1, 2, 10) / np.e
    ]
    fraction_non_clifford = 0.4

    indices = _select(
        non_clifford_ops,
        fraction_non_clifford,
        method=method,
        sigma=0.5,
        random_state=np.random.RandomState(1),
    )
    assert all(isinstance(index, int) for index in indices)
    assert len(indices) == int(
        round((1.0 - fraction_non_clifford) * len(non_clifford_ops))
    )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import functools

import cirq
import numpy as np
import pytest

from mitiq import MeasurementResult
from mitiq.interface.mitiq_cirq.cirq_utils import (
    compute_density_matrix,
    sample_bitstrings,
)
from mitiq.observable.observable import Observable
from mitiq.observable.pauli import PauliString, PauliStringCollection
from mitiq.utils import _equal

# Basis rotations to measure Pauli X and Y.
xrotation = cirq.SingleQubitCliffordGate.Y_nsqrt
yrotation = cirq.SingleQubitCliffordGate.X_sqrt

# Pauli matrices.
imat = np.identity(2)
xmat = cirq.unitary(cirq.X)
zmat = cirq.unitary(cirq.Z)


def test_observable():
    pauli1 = PauliString(spec="XI", coeff=-1.0)
    pauli2 = PauliString(spec="IZ", coeff=2.0)
    obs = Observable(pauli1, pauli2)

    assert obs.nqubits == 2
    assert obs.qubit_indices == [0, 1]
    assert obs.nterms == 2
    assert obs.ngroups == 1

    assert str(obs) == "-X(q(0)) + (2+0j)*Z(q(1))"

    correct_matrix = -1.0 * np.kron(xmat, imat) + 2.0 * np.kron(imat, zmat)
    assert np.allclose(obs.matrix(), correct_matrix)


def test_observable_from_pauli_string_collections():
    z = PauliString("Z")
    zz = PauliString("ZZ")
    pcol1 = PauliStringCollection(z, zz)
    x = PauliString("X")
    xx = PauliString("XX")
    pcol2 = PauliStringCollection(x, xx)

    obs = Observable.from_pauli_string_collections(pcol1, pcol2)
    assert obs.ngroups == 2
    assert obs.groups[0] == pcol1
    assert obs.groups[1] == pcol2
    assert obs.nterms == 4


def test_observable_partition_one_set():
    pauli1 = PauliString(spec="ZI")
    pauli2 = PauliString(spec="IZ")
    pauli3 = PauliString(spec="ZZ")
    obs = Observable(pauli1, pauli2, pauli3)
    assert obs.nterms == 3

    assert obs.ngroups == 1
    assert obs.groups[0] == PauliStringCollection(pauli1, pauli2, pauli3)


def test_observable_partition_single_qubit_paulis():
    x = PauliString(spec="X")
    y = PauliString(spec="Y")
    z = PauliString(spec="Z")
    obs = Observable(x, y, z)
    assert obs.nterms == 3

    obs.partition(seed=2)
    expected_groups = [
        PauliStringCollection(x),
        PauliStringCollection(y),
        PauliStringCollection(z),
    ]
    assert obs.groups == expected_groups


def test_observable_partition_can_be_measured_with():
    n = 10
    nterms = 50
    rng = np.random.RandomState(seed=1)
    obs = Observable(
        *[
            PauliString(
                spec=rng.choice(
                    ("I", "X", "Y", "Z"),
                    size=n,
                    replace=True,
                    p=(0.7, 0.1, 0.1, 0.1),
                )
            )
            for _ in range(nterms)
        ]
    )

    assert obs.nqubits == n
    assert obs.nterms <= nterms  # because of deduplication
    assert obs.ngroups <= obs.nterms

    for pset in obs.groups:
        pauli_list = list(pset.elements)
        for i in range(len(pauli_list) - 1):
            for j in range(i, len(pauli_list)):
                assert pauli_list[i].can_be_measured_with(pauli_list[j])


def test_observable_measure_in_needs_one_circuit_z():
    pauli1 = PauliString(spec="ZI")
    pauli2 = PauliString(spec="IZ")
    pauli3 = PauliString(spec="ZZ")
    obs = Observable(pauli1, pauli2, pauli3)

    qubits = cirq.LineQubit.range(2)
    circuit = cirq.testing.random_circuit(qubits, 3, 1, random_state=1)

    measures_obs_circuits = obs.measure_in(circuit)
    assert len(measures_obs_circuits) == 1

    expected = circuit + cirq.measure(*qubits)
    assert _equal(
        measures_obs_circuits[0],
        expected,
        require_qubit_equality=True,
        require_measurement_equality=True,
    )


def test_observable_measure_in_needs_one_circuit_x():
    pauli1 = PauliString(spec="XI")
    pauli2 = PauliString(spec="IX")
    pauli3 = PauliString(spec="XX")
    obs = Observable(pauli1, pauli2, pauli3)

    qubits = cirq.LineQubit.range(2)
    circuit = cirq.testing.random_circuit(qubits, 3, 1, random_state=1)

    measures_obs_circuits = obs.measure_in(circuit)
    assert len(measures_obs_circuits) == 1

    expected = circuit + xrotation.on_each(*qubits) + cirq.measure(*qubits)
    assert _equal(
        measures_obs_circuits[0],
        expected,
        require_qubit_equality=True,
        require_measurement_equality=True,
    )


def test_observable_measure_in_needs_two_circuits():
    obs = Observable(PauliString(spec="X"), PauliString(spec="Z"))

    q = cirq.LineQubit(0)
    circuit = cirq.Circuit(cirq.H.on(q))

    measures_obs_circuits = sorted(obs.measure_in(circuit), key=len)
    assert len(measures_obs_circuits) == 2

    expected_circuits = [
        circuit + cirq.measure(q),
        circuit + xrotation.on(q) + cirq.measure(q),
    ]
    for expected, measured in zip(expected_circuits, measures_obs_circuits):
        assert _equal(
            measured,
            expected,
            require_qubit_equality=True,
            require_measurement_equality=True,
        )


def test_observable_expectation_from_measurements_one_pauli_string():
    obs = Observable(PauliString(spec="Z"))

    measurements = MeasurementResult(
        [[0], [0], [0], [0], [0], [0], [0], [0], [0], [0]]
    )
    expectation = obs._expectation_from_measurements([measurements])
    assert np.isclose(expectation, 1.0)

    measurements = MeasurementResult(
        [[1], [1], [1], [1], [1], [1], [1], [1], [1], [1]]
    )
    expectation = obs._expectation_from_measurements([measurements])
    assert np.isclose(expectation, -1.0)

    measurements = MeasurementResult(
        [[0], [1], [0], [1], [0], [1], [0], [1], [0], [1]]
    )
    expectation = obs._expectation_from_measurements([measurements])
    assert np.isclose(expectation, 0.0)


def test_observable_expectation_from_measurements_two_pauli_strings():
    obs = Observable(PauliString(spec="Z", coeff=2.5), PauliString(spec="X"))

    bits = MeasurementResult(
        [[0], [0], [0], [0], [0], [0], [0], [0], [0], [0]]
    )
    expectation = obs._expectation_from_measurements([bits, bits])
    assert np.isclose(expectation, 3.5)


@pytest.mark.parametrize("n", range(1, 3 + 1))
@pytest.mark.parametrize(
    "executor", (sample_bitstrings, compute_density_matrix)
)
def test_observable_expectation_one_circuit(n, executor):
    executor = functools.partial(executor, noise_level=(0,))

    qubits = cirq.LineQubit.range(n)
    obs = Observable(PauliString(spec="X" * n))
    circuit = cirq.Circuit(cirq.H.on_each(qubits))

    expectation = obs.expectation(circuit, executor)
    assert np.isclose(expectation, 1.0)


@pytest.mark.parametrize("n", range(1, 3 + 1))
@pytest.mark.parametrize(
    "executor", (sample_bitstrings, compute_density_matrix)
)
def test_observable_expectation_two_circuits(n, executor):
    executor = functools.partial(executor, noise_level=(0,))

    obs = Observable(
        PauliString(spec="X" * n, coeff=-2.0), PauliString(spec="Z" * n)
    )
    qubits = cirq.LineQubit.range(n)
    circuit = cirq.Circuit(cirq.H.on_each(qubits))

    expectation = obs.expectation(circuit, executor)
    assert np.isclose(expectation, -2.0, atol=1e-1)


@pytest.mark.parametrize(
    "executor", (sample_bitstrings, compute_density_matrix)
)
def test_observable_expectation_supported_qubits(executor):
    executor = functools.partial(executor, noise_level=(0,))

    a, b, c = cirq.LineQubit.range(3)
    circuit = cirq.Circuit(cirq.I(a), cirq.X.on(b), cirq.H.on(c))

    # <Z0> = 1.
    obs = Observable(PauliString(spec="Z", support=(0,)))
    assert np.isclose(obs.expectation(circuit, executor), 1.0)

    # <Z1> = -1.
    obs = Observable(PauliString(spec="Z", support=(1,)))
    assert np.isclose(obs.expectation(circuit, executor), -1.0)

    # <Z2> = 0.
    obs = Observable(PauliString(spec="Z", support=(2,)))
    assert np.isclose(obs.expectation(circuit, executor), 0.0, atol=5e-2)


def test_observable_multuplication_1():
    XI = PauliString("XI", 0.3)
    YY = PauliString("YY", 0.7)
    XZ = PauliString("XZ", 0.1)
    ZX = PauliString("ZX", 0.2)
    IZ = PauliString("IZ", -0.4)
    obs1 = Observable(XI, YY, XZ)
    obs2 = Observable(ZX, IZ)
    correct_obs = Observable(
        XI * ZX, XI * IZ, YY * ZX, YY * IZ, XZ * ZX, XZ * IZ
    )
    assert obs1 * obs2 == correct_obs


def test_observable_multiplication_2():
    YXXYZ = PauliString("YXXYZ", 0.3)
    ZYIZX = PauliString("ZYIZX", 0.7)
    IZZXY = PauliString("IZZXY", 0.1)
    YZIXZ = PauliString("YZIXZ", 0.2)
    XZYII = PauliString("XZYII", -0.4)
    YYZXI = PauliString("YYZXI", 0.7)
    IIXYZ = PauliString("IIXYZ", 0.7)
    ZYXIZ = PauliString("ZYXIZ", 0.1)
    YIZXI = PauliString("YIZXI", 0.2)
    pauli_strings_1 = [YXXYZ, ZYIZX, IZZXY, YZIXZ, XZYII]
    pauli_strings_2 = [YYZXI, IIXYZ, ZYXIZ, YIZXI]
    obs1 = Observable(*pauli_strings_1)
    obs2 = Observable(*pauli_strings_2)
    l3 = [p1 * p2 for p1 in pauli_strings_1 for p2 in pauli_strings_2]
    correct_obs = Observable(*l3)
    assert obs1 * obs2 == correct_obs


def test_scalar_multiplication():
    YXXYZ = PauliString("YXXYZ", 0.3)
    obs = Observable(YXXYZ)
    assert obs * 2.0 == Observable(YXXYZ * 2.0)
    assert 2.0 * obs == Observable(YXXYZ * 2.0)


def test_pauli_string_left_multiplication():
    XI = PauliString("XI", 0.3)
    YY = PauliString("YY", 0.7)
    XZ = PauliString("XZ", 0.1)
    IZ = PauliString("IZ", -0.4)
    pauli_strings = [XI, YY, XZ]
    obs1 = Observable(*pauli_strings)
    correct_obs = Observable(*[p * IZ for p in pauli_strings])
    assert obs1 * IZ == correct_obs


def test_pauli_string_right_multiplication():
    XI = PauliString("XI", 0.3)
    YY = PauliString("YY", 0.7)
    XZ = PauliString("XZ", 0.1)
    IZ = PauliString("IZ", -0.4)
    pauli_strings = [XI, YY, XZ]
    obs1 = Observable(*pauli_strings)
    correct_obs = Observable(*[IZ * p for p in pauli_strings])
    assert IZ * obs1 == correct_obs


def test_pauli_string_deduplication():
    XI = PauliString("XI", 0.3)
    YY = PauliString("YY", 0.7)
    XZ = PauliString("XZ", 0.1)
    n_repeat_paulis = 4
    pauli_strings = [XI, YY, XZ] * n_repeat_paulis

    obs = Observable(*pauli_strings)
    assert obs.nterms == 3
    assert obs == Observable(
        XI.with_coeff(XI.coeff * n_repeat_paulis),
        YY.with_coeff(YY.coeff * n_repeat_paulis),
        XZ.with_coeff(XZ.coeff * n_repeat_paulis),
    )


def test_pauli_string_deduplication_removal_of_0_coefficients():
    XI = PauliString("XI", 0.3)
    YY = PauliString("YY", 0.7)

    pauli_strings = [XI, YY, XI, YY.with_coeff(-YY.coeff)]

    obs = Observable(*pauli_strings)
    assert obs.nterms == 1
    assert obs == Observable(XI.with_coeff(XI.coeff * 2))



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for quantum volume circuits. The Cirq functions that do the main work
are tested here:
cirq-core/cirq/contrib/quantum_volume/quantum_volume_test.py

Tests below check that generate_quantum_volume_circuit() works as a wrapper and
fits with Mitiq's interface.
"""

import cirq
import pytest
from cirq import protocols

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks.quantum_volume_circuits import (
    compute_heavy_bitstrings,
    generate_quantum_volume_circuit,
)


def test_generate_model_circuit_no_seed():
    """Test that random circuit of the right length is generated."""
    circuit, _ = generate_quantum_volume_circuit(3, 3)
    assert len(circuit) == 3


def test_generate_model_circuit_with_seed():
    """Test that a model circuit is determined by its seed."""
    circuit_1, _ = generate_quantum_volume_circuit(3, 3, seed=1)
    circuit_2, _ = generate_quantum_volume_circuit(3, 3, seed=1)
    circuit_3, _ = generate_quantum_volume_circuit(3, 3, seed=2)

    assert circuit_1 == circuit_2
    assert circuit_2 != circuit_3


def test_compute_heavy_bitstrings():
    """Test that the heavy bitstrings can be computed from a given circuit."""
    a, b, c = cirq.LineQubit.range(3)
    model_circuit = cirq.Circuit(
        [
            cirq.Moment([]),
            cirq.Moment([cirq.X(a), cirq.Y(b)]),
            cirq.Moment([]),
            cirq.Moment([cirq.CNOT(a, c)]),
            cirq.Moment([cirq.Z(a), cirq.H(b)]),
        ]
    )
    true_heavy_set = [[1, 0, 1], [1, 1, 1]]
    computed_heavy_set = compute_heavy_bitstrings(model_circuit, 3)
    assert computed_heavy_set == true_heavy_set


def test_circuit_decomposition():
    """Test that decomposed circuit consists of gates in default cirq gatest.
    As defined in cirq.protocols.decompose_protocol, this default gateset is
        ops.XPowGate,
        ops.YPowGate,
        ops.ZPowGate,
        ops.CZPowGate,
        ops.MeasurementGate,
        ops.GlobalPhaseGate
    """
    circuit, _ = generate_quantum_volume_circuit(3, 3, decompose=True)
    for op in [operation for moment in circuit for operation in moment]:
        assert op in protocols.decompose_protocol.DECOMPOSE_TARGET_GATESET


@pytest.mark.parametrize("return_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_volume_conversion(return_type):
    circuit, _ = generate_quantum_volume_circuit(3, 3, return_type=return_type)
    assert return_type in circuit.__module__



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for utility functions."""

from copy import deepcopy

import cirq
import numpy as np
import pytest
from cirq import (
    CNOT,
    Circuit,
    ControlledGate,
    H,
    LineQubit,
    MeasurementGate,
    S,
    T,
    X,
    Y,
    Z,
    depolarize,
    ops,
)

from mitiq.utils import (
    _append_measurements,
    _are_close_dict,
    _circuit_to_choi,
    _equal,
    _is_measurement,
    _max_ent_state_circuit,
    _operation_to_choi,
    _pop_measurements,
    _simplify_circuit_exponents,
    _simplify_gate_exponent,
    arbitrary_tensor_product,
    matrix_kronecker_product,
    matrix_to_vector,
    operator_ptm_vector_rep,
    qem_methods,
    vector_to_matrix,
)


def test_arbitrary_tensor_product():
    terms = [np.random.rand(dim, dim) for dim in range(1, 4)]
    expected = np.kron(np.kron(terms[0], terms[1]), terms[2])
    assert np.allclose(arbitrary_tensor_product(*terms), expected)
    # Check limit cases
    one_term = np.random.rand(5, 5)
    assert np.allclose(arbitrary_tensor_product(one_term), one_term)
    assert np.allclose(arbitrary_tensor_product(2.0, one_term), 2.0 * one_term)
    assert np.allclose(arbitrary_tensor_product(3.0, 4.0), 12.0)
    with pytest.raises(TypeError, match="requires at least one argument"):
        assert np.allclose(arbitrary_tensor_product(), one_term)


def test_matrix_to_vector():
    for d in [1, 2, 3, 4]:
        mat = np.random.rand(d, d)
        assert matrix_to_vector(mat).shape == (d**2,)
        assert (vector_to_matrix(matrix_to_vector(mat)) == mat).all


def test_vector_to_matrix():
    for d in [1, 2, 3, 4]:
        vec = np.random.rand(d**2)
        assert vector_to_matrix(vec).shape == (d, d)
        assert (matrix_to_vector(vector_to_matrix(vec)) == vec).all


@pytest.mark.parametrize("require_qubit_equality", [True, False])
def test_circuit_equality_identical_qubits(require_qubit_equality):
    qreg = cirq.NamedQubit.range(5, prefix="q_")
    circA = cirq.Circuit(cirq.ops.H.on_each(*qreg))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qreg))
    assert circA is not circB
    assert _equal(circA, circB, require_qubit_equality=require_qubit_equality)


@pytest.mark.parametrize("require_qubit_equality", [True, False])
def test_circuit_equality_nonidentical_but_equal_qubits(
    require_qubit_equality,
):
    n = 5
    qregA = cirq.NamedQubit.range(n, prefix="q_")
    qregB = cirq.NamedQubit.range(n, prefix="q_")
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert circA is not circB
    assert _equal(circA, circB, require_qubit_equality=require_qubit_equality)


def test_circuit_equality_linequbit_gridqubit_equal_indices():
    n = 10
    qregA = cirq.LineQubit.range(n)
    qregB = [cirq.GridQubit(x, 0) for x in range(n)]
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert _equal(circA, circB, require_qubit_equality=False)
    assert not _equal(circA, circB, require_qubit_equality=True)


def test_circuit_equality_linequbit_gridqubit_unequal_indices():
    n = 10
    qregA = cirq.LineQubit.range(n)
    qregB = [cirq.GridQubit(x + 3, 0) for x in range(n)]
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert _equal(circA, circB, require_qubit_equality=False)
    assert not _equal(circA, circB, require_qubit_equality=True)


def test_circuit_equality_linequbit_namedqubit_equal_indices():
    n = 8
    qregA = cirq.LineQubit.range(n)
    qregB = cirq.NamedQubit.range(n, prefix="q_")
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert _equal(circA, circB, require_qubit_equality=False)
    assert not _equal(circA, circB, require_qubit_equality=True)


def test_circuit_equality_linequbit_namedqubit_unequal_indices():
    n = 11
    qregA = cirq.LineQubit.range(n)
    qregB = [cirq.NamedQubit(str(x + 10)) for x in range(n)]
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert _equal(circA, circB, require_qubit_equality=False)
    assert not _equal(circA, circB, require_qubit_equality=True)


def test_circuit_equality_gridqubit_namedqubit_equal_indices():
    n = 8
    qregA = [cirq.GridQubit(0, x) for x in range(n)]
    qregB = cirq.NamedQubit.range(n, prefix="q_")
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert _equal(circA, circB, require_qubit_equality=False)
    assert not _equal(circA, circB, require_qubit_equality=True)


def test_circuit_equality_gridqubit_namedqubit_unequal_indices():
    n = 5
    qregA = [cirq.GridQubit(x + 2, 0) for x in range(n)]
    qregB = [cirq.NamedQubit(str(x + 10)) for x in range(n)]
    circA = cirq.Circuit(cirq.ops.H.on_each(*qregA))
    circB = cirq.Circuit(cirq.ops.H.on_each(*qregB))
    assert _equal(circA, circB, require_qubit_equality=False)
    assert not _equal(circA, circB, require_qubit_equality=True)


def test_circuit_equality_unequal_measurement_keys_terminal_measurements():
    base_circuit = cirq.testing.random_circuit(
        qubits=5, n_moments=10, op_density=0.99, random_state=1
    )
    qreg = list(base_circuit.all_qubits())

    circ1 = deepcopy(base_circuit)
    circ1.append(cirq.measure(q, key="one") for q in qreg)

    circ2 = deepcopy(base_circuit)
    circ2.append(cirq.measure(q, key="two") for q in qreg)

    assert _equal(circ1, circ2, require_measurement_equality=False)
    assert not _equal(circ1, circ2, require_measurement_equality=True)


@pytest.mark.parametrize("require_measurement_equality", [True, False])
def test_circuit_equality_equal_measurement_keys_terminal_measurements(
    require_measurement_equality,
):
    base_circuit = cirq.testing.random_circuit(
        qubits=5, n_moments=10, op_density=0.99, random_state=1
    )
    qreg = list(base_circuit.all_qubits())

    circ1 = deepcopy(base_circuit)
    circ1.append(cirq.measure(q, key="z") for q in qreg)

    circ2 = deepcopy(base_circuit)
    circ2.append(cirq.measure(q, key="z") for q in qreg)

    assert _equal(
        circ1, circ2, require_measurement_equality=require_measurement_equality
    )


def test_circuit_equality_unequal_measurement_keys_nonterminal_measurements():
    base_circuit = cirq.testing.random_circuit(
        qubits=5, n_moments=10, op_density=0.99, random_state=1
    )
    end_circuit = cirq.testing.random_circuit(
        qubits=5, n_moments=5, op_density=0.99, random_state=2
    )
    qreg = list(base_circuit.all_qubits())

    circ1 = deepcopy(base_circuit)
    circ1.append(cirq.measure(q, key="one") for q in qreg)
    circ1 += end_circuit

    circ2 = deepcopy(base_circuit)
    circ2.append(cirq.measure(q, key="two") for q in qreg)
    circ2 += end_circuit

    assert _equal(circ1, circ2, require_measurement_equality=False)
    assert not _equal(circ1, circ2, require_measurement_equality=True)


@pytest.mark.parametrize("require_measurement_equality", [True, False])
def test_circuit_equality_equal_measurement_keys_nonterminal_measurements(
    require_measurement_equality,
):
    base_circuit = cirq.testing.random_circuit(
        qubits=5, n_moments=10, op_density=0.99, random_state=1
    )
    end_circuit = cirq.testing.random_circuit(
        qubits=5, n_moments=5, op_density=0.99, random_state=2
    )
    qreg = list(base_circuit.all_qubits())

    circ1 = deepcopy(base_circuit)
    circ1.append(cirq.measure(q, key="z") for q in qreg)
    circ1 += end_circuit

    circ2 = deepcopy(base_circuit)
    circ2.append(cirq.measure(q, key="z") for q in qreg)
    circ2 += end_circuit

    assert _equal(
        circ1, circ2, require_measurement_equality=require_measurement_equality
    )


def test_is_measurement():
    """Tests for checking if operations are measurements."""
    # Test circuit:
    # 0: ───H───X───Z───
    qbit = LineQubit(0)
    circ = Circuit(
        [ops.H.on(qbit), ops.X.on(qbit), ops.Z.on(qbit), ops.measure(qbit)]
    )
    for i, op in enumerate(circ.all_operations()):
        if i == 3:
            assert _is_measurement(op)
        else:
            assert not _is_measurement(op)


def test_pop_measurements_and_add_measurements():
    """Tests popping measurements from a circuit.."""
    # Test circuit:
    # 0: ───H───T───@───M───
    #               │   │
    # 1: ───H───M───┼───┼───
    #               │   │
    # 2: ───H───────X───M───
    qreg = LineQubit.range(3)
    circ = Circuit(
        [ops.H.on_each(qreg)],
        [ops.T.on(qreg[0])],
        [ops.measure(qreg[1])],
        [ops.CNOT.on(qreg[0], qreg[2])],
        [ops.measure(qreg[0], qreg[2])],
    )
    copy = deepcopy(circ)
    measurements = _pop_measurements(copy)
    correct = Circuit(
        [ops.H.on_each(qreg)],
        [ops.T.on(qreg[0])],
        [ops.CNOT.on(qreg[0], qreg[2])],
    )
    assert _equal(copy, correct)
    _append_measurements(copy, measurements)
    assert _equal(copy, circ)


@pytest.mark.parametrize("gate", [X**3, Y**-3, Z**-1, H**-1])
def test_simplify_gate_exponent(gate):
    # Check exponent is simplified to 1
    assert _simplify_gate_exponent(gate).exponent == 1
    # Check simplified gate is equivalent to the input
    assert _simplify_gate_exponent(gate) == gate


@pytest.mark.parametrize("gate", [T**-1, S**-1, MeasurementGate(1)])
def test_simplify_gate_exponent_with_gates_that_cannot_be_simplified(gate):
    # Check the gate is not simplified (same representation)
    assert _simplify_gate_exponent(gate).__repr__() == gate.__repr__()


def test_simplify_circuit_exponents_controlled_gate():
    circuit = Circuit(
        ControlledGate(CNOT, num_controls=1).on(*LineQubit.range(3))
    )
    copy = circuit.copy()

    _simplify_circuit_exponents(circuit)
    assert _equal(circuit, copy)


def test_simplify_circuit_exponents():
    qreg = LineQubit.range(2)
    circuit = Circuit([H.on(qreg[0]), CNOT.on(*qreg), Z.on(qreg[1])])

    # Invert circuit
    inverse_circuit = cirq.inverse(circuit)
    inverse_repr = inverse_circuit.__repr__()
    inverse_qasm = inverse_circuit._to_qasm_output().__str__()

    # Expected circuit after simplification
    expected_inv = Circuit([Z.on(qreg[1]), CNOT.on(*qreg), H.on(qreg[0])])
    expected_repr = expected_inv.__repr__()
    expected_qasm = expected_inv._to_qasm_output().__str__()

    # Check inverse_circuit is logically equivalent to expected_inverse
    # but they have a different representation
    assert inverse_circuit == expected_inv
    assert inverse_repr != expected_repr
    assert inverse_qasm != expected_qasm

    # Simplify the circuit
    _simplify_circuit_exponents(inverse_circuit)

    # Check inverse_circuit has the expected simplified representation
    simplified_repr = inverse_circuit.__repr__()
    simplified_qasm = inverse_circuit._to_qasm_output().__str__()
    assert inverse_circuit == expected_inv
    assert simplified_repr == expected_repr
    assert simplified_qasm == expected_qasm


def test_simplify_circuit_exponents_with_non_self_inverse_gates():
    qreg = LineQubit.range(2)
    # Make a circuit with gates which are not self-inverse
    circuit = Circuit([S.on(qreg[0]), T.on(qreg[1])])

    inverse_circuit = cirq.inverse(circuit)
    inverse_repr = inverse_circuit.__repr__()
    inverse_qasm = inverse_circuit._to_qasm_output().__str__()

    # Simplify the circuit (it should not change this circuit)
    _simplify_circuit_exponents(inverse_circuit)

    # Check inverse_circuit did not change
    simplified_repr = inverse_circuit.__repr__()
    simplified_qasm = inverse_circuit._to_qasm_output().__str__()
    assert simplified_repr == inverse_repr
    assert simplified_qasm == inverse_qasm


def test_are_close_dict():
    """Tests the _are_close_dict function."""
    dict1 = {"a": 1, "b": 0.0}
    dict2 = {"a": 1, "b": 0.0 + 1.0e-10}
    assert _are_close_dict(dict1, dict2)
    assert _are_close_dict(dict2, dict1)
    dict2 = {"b": 0.0 + 1.0e-10, "a": 1}
    assert _are_close_dict(dict1, dict2)
    assert _are_close_dict(dict2, dict1)
    dict2 = {"a": 1, "b": 1.0}
    assert not _are_close_dict(dict1, dict2)
    assert not _are_close_dict(dict2, dict1)
    dict2 = {"b": 1, "a": 0.0}
    assert not _are_close_dict(dict1, dict2)
    assert not _are_close_dict(dict2, dict1)
    dict2 = {"a": 1, "b": 0.0, "c": 1}
    assert not _are_close_dict(dict1, dict2)
    assert not _are_close_dict(dict2, dict1)


def test_max_ent_state_circuit():
    """Tests 1-qubit and 2-qubit maximally entangled states are generated."""
    two_state = np.array([1, 0, 0, 1]) / np.sqrt(2)
    four_state = np.array(3 * [1, 0, 0, 0, 0] + [1]) / 2.0
    assert np.allclose(
        _max_ent_state_circuit(2).final_state_vector(), two_state
    )
    assert np.allclose(
        _max_ent_state_circuit(4).final_state_vector(), four_state
    )


def test_circuit_to_choi_and_operation_to_choi():
    """Tests the Choi matrix of a depolarizing channel is recovered."""
    # Define first the expected result
    base_noise = 0.01
    max_ent_state = np.array([1, 0, 0, 1]) / np.sqrt(2)
    identity_part = np.outer(max_ent_state, max_ent_state)
    mixed_part = np.eye(4) / 4.0
    epsilon = base_noise * 4.0 / 3.0
    choi = (1.0 - epsilon) * identity_part + epsilon * mixed_part
    # Choi matrix of the double application of a depolarizing channel
    choi_twice = sum(
        [
            ((1.0 - epsilon) ** 2 * identity_part),
            (2 * epsilon - epsilon**2) * mixed_part,
        ]
    )

    # Evaluate the Choi matrix of one or two depolarizing channels
    q = LineQubit(0)
    noisy_operation = depolarize(base_noise).on(q)
    noisy_sequence = [noisy_operation, noisy_operation]
    assert np.allclose(choi, _operation_to_choi(noisy_operation))
    noisy_circuit = Circuit(noisy_operation)
    noisy_circuit_twice = Circuit(noisy_sequence)
    assert np.allclose(choi, _circuit_to_choi(noisy_circuit))
    assert np.allclose(choi_twice, _circuit_to_choi(noisy_circuit_twice))


def test_kronecker_product():
    matrices = [np.array([[1, 2], [3, 4]]), np.array([[0, 1], [1, 0]])]
    expected_result = np.array(
        [[0, 1, 0, 2], [1, 0, 2, 0], [0, 3, 0, 4], [3, 0, 4, 0]]
    )
    np.testing.assert_array_equal(
        matrix_kronecker_product(matrices), expected_result
    )


def test_operator_ptm_vector_rep():
    opt = cirq.I._unitary_() / np.sqrt(2)
    expected_result = np.array([1.0, 0.0, 0.0, 0.0])
    np.testing.assert_array_almost_equal(
        operator_ptm_vector_rep(opt), expected_result
    )


def test_operator_ptm_vector_rep_raised_error():
    with pytest.raises(TypeError, match="Input must be a square matrix"):
        assert np.allclose(
            operator_ptm_vector_rep(np.array([1.0, 0.0, 0.0, 0.0]))
        )


def test_qem_methods_basic():
    for module, name in qem_methods().items():
        prefix, suffix = module.split(".")
        assert prefix == "mitiq"
        assert len(suffix) <= 3



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from collections import Counter

import cirq
import numpy as np
import pytest

from mitiq import MeasurementResult
from mitiq.interface import mitiq_pyquil, mitiq_qiskit
from mitiq.observable.pauli import PauliString, PauliStringCollection
from mitiq.utils import _equal

# Basis rotations to measure Pauli X and Y.
xrotation = cirq.SingleQubitCliffordGate.Y_nsqrt
yrotation = cirq.SingleQubitCliffordGate.X_sqrt

# Matrices.
imat = np.identity(2)
xmat = cirq.unitary(cirq.X)
zmat = cirq.unitary(cirq.Z)


def test_pauli_init():
    pauli = PauliString(spec="IZXYI", coeff=1.0)
    a, b, c = cirq.LineQubit.range(1, 4)
    assert pauli._pauli == cirq.PauliString(
        1.0, cirq.Z(a), cirq.X(b), cirq.Y(c)
    )
    assert str(pauli) == "Z(q(1))*X(q(2))*Y(q(3))"


def test_pauli_init_empty():
    pauli = PauliString()
    assert pauli.support() == set()
    assert pauli.weight() == 0
    assert np.allclose(pauli.matrix(), [[1]])


def test_pauli_init_with_support():
    support = (2, 37)
    pauli = PauliString(spec="XZ", support=support)

    assert pauli._pauli == cirq.PauliString(
        cirq.X(cirq.LineQubit(support[0])), cirq.Z(cirq.LineQubit(support[1]))
    )
    assert str(pauli) == f"X(q({support[0]}))*Z(q({support[1]}))"


def test_pauli_eq():
    assert PauliString(spec="Z") == PauliString(spec="Z")
    assert PauliString(spec="X") != PauliString(spec="Z")
    assert PauliString(spec="Z") != PauliString(spec="Z", coeff=-1.0)

    assert PauliString(spec="Z", support=(0,)) != PauliString(
        spec="Z", support=(1,)
    )
    assert PauliString(spec="IZ") == PauliString(spec="Z", support=(1,))
    assert PauliString(spec="XY") == PauliString(spec="YX", support=(1, 0))

    assert {PauliString(spec="Z"), PauliString(spec="Z")} == {
        PauliString(spec="Z")
    }


def test_matrix():
    assert np.allclose(PauliString(spec="X").matrix(), xmat)
    assert np.allclose(PauliString(spec="Z", coeff=-0.5).matrix(), -0.5 * zmat)
    assert np.allclose(PauliString(spec="ZZ").matrix(), np.kron(zmat, zmat))
    assert np.allclose(PauliString(spec="XZ").matrix(), np.kron(xmat, zmat))


def test_pauli_matrix_include_qubits():
    pauli = PauliString(spec="X")

    assert np.allclose(pauli.matrix(), xmat)
    assert np.allclose(
        pauli.matrix(qubit_indices_to_include=[0, 1]), np.kron(xmat, imat)
    )
    assert np.allclose(
        pauli.matrix(qubit_indices_to_include=[0, 1, 2]),
        np.kron(np.kron(xmat, imat), imat),
    )


@pytest.mark.parametrize("support", [range(3), range(1, 4)])
@pytest.mark.parametrize("circuit_type", ("cirq", "qiskit", "pyquil"))
def test_pauli_measure_in_circuit(support, circuit_type):
    pauli = PauliString(spec="XYZ", support=support, coeff=-0.5)

    names = ("0th", "1st", "2nd", "3rd", "4th", "5th")
    qreg = [cirq.NamedQubit(name) for name in names]
    base_circuit = cirq.Circuit(cirq.H.on_each(qreg))

    if circuit_type == "cirq":

        def convert(circ):
            return circ

    elif circuit_type == "qiskit":
        convert = mitiq_qiskit.to_qiskit
    elif circuit_type == "pyquil":
        convert = mitiq_pyquil.to_pyquil

    circuit = convert(base_circuit)
    measured = pauli.measure_in(circuit)

    qreg = [cirq.NamedQubit(name) for name in names]
    expected = cirq.Circuit(
        # Original circuit.
        base_circuit.all_operations(),
        # Basis rotations.
        xrotation.on(qreg[support[0]]),
        yrotation.on(qreg[support[1]]),
        # Measurements.
        cirq.measure(*[qreg[s] for s in support]),
    )
    if circuit_type == "cirq":
        assert _equal(measured, expected, require_qubit_equality=True)
    else:
        expected = convert(expected)
        if circuit_type == "pyquil":  # Special case with basis rotation order.
            assert set(measured) == set(expected)
        else:
            assert measured == expected


def test_pauli_measure_in_bad_qubits_error():
    n = 5
    pauli = PauliString(spec="X" * n)
    circuit = cirq.Circuit(cirq.H.on_each(cirq.LineQubit.range(n - 1)))

    with pytest.raises(ValueError, match="Qubit mismatch."):
        pauli.measure_in(circuit)


def test_pauli_measure_in_multi__measurement_per_qubit():
    n = 4
    pauli = PauliString(spec="Z" * n)
    circuit = cirq.Circuit(cirq.H.on_each(cirq.LineQubit.range(n)))

    # add a measurement to qubit 0 and 1
    circuit = circuit + cirq.measure(cirq.LineQubit(0), cirq.LineQubit(1))
    with pytest.raises(ValueError, match="More than one measurement"):
        pauli.measure_in(circuit)


def test_can_be_measured_with_single_qubit():
    pauli = PauliString(spec="Z")

    assert pauli.can_be_measured_with(PauliString(spec="I"))
    assert not pauli.can_be_measured_with(PauliString(spec="X"))
    assert not pauli.can_be_measured_with(PauliString(spec="Y"))
    assert pauli.can_be_measured_with(PauliString(spec="Z", coeff=-0.5))
    assert pauli.can_be_measured_with(pauli)


def test_can_be_measured_with_two_qubits():
    pauli = PauliString(spec="ZX")

    assert pauli.can_be_measured_with(PauliString(spec="Z"))
    assert pauli.can_be_measured_with(PauliString(spec="X", support=(1,)))
    assert not pauli.can_be_measured_with(PauliString(spec="X"))
    assert not pauli.can_be_measured_with(PauliString(spec="Y", coeff=-1.0))

    assert pauli.can_be_measured_with(PauliString(spec="ZX", coeff=0.5))
    assert not pauli.can_be_measured_with(PauliString(spec="ZZ"))


def test_can_be_measured_with_non_overlapping_paulis():
    pauli = PauliString(spec="ZX")

    assert pauli.can_be_measured_with(PauliString(spec="IIZI"))
    assert pauli.can_be_measured_with(PauliString(spec="IIIX"))
    assert pauli.can_be_measured_with(PauliString(spec="IIYZ"))


def test_weight():
    assert PauliString(spec="I").weight() == 0
    assert PauliString(spec="Z").weight() == 1

    n = 4
    assert PauliString(spec="X" * n).weight() == n
    assert PauliString(spec="IX" * n).weight() == n
    assert PauliString(spec="ZX" * n).weight() == 2 * n


def test_multiplication():
    Pauli = PauliString
    assert 2 * Pauli("X") == Pauli("X", coeff=2)
    assert Pauli("X") * 2 == Pauli("X", coeff=2)
    assert Pauli("X") * Pauli("I") == Pauli("X")
    assert Pauli("X") * Pauli("Y") == Pauli("Z", coeff=1j)
    assert Pauli("X") * Pauli("Y", support=(1,)) == Pauli("XY")
    assert Pauli("ZI", coeff=2) * Pauli("IZ", coeff=3) == Pauli("ZZ", coeff=6)

    zz_mult = Pauli("ZI") * Pauli("IZ")
    zz_expected = Pauli("ZZ")
    assert zz_mult.support() == zz_expected.support()
    assert np.allclose(zz_mult.matrix(), zz_expected.matrix())
    assert str(zz_mult) == str(zz_expected)
    assert zz_mult.weight() == zz_expected.weight()


# Note: For testing `PauliString._expectation_from_measurements`, it makes no
# difference whether the Pauli is X, Y, or Z. This is because we assume
# measurements are obtained by single-qubit basis rotations. So it only matters
# whether the Pauli is I (identity) or X, Y, Z (not identity). We just use Z
# for "not identity" below.


@pytest.mark.parametrize("seed", range(5))
@pytest.mark.parametrize("nqubits", [1, 2, 5])
def test_expectation_from_measurements_identity(seed, nqubits):
    """For P = cI, asserts ⟨P⟩ = c."""
    rng = np.random.RandomState(seed)
    coeff = rng.random()
    pauli = PauliString(spec="I", coeff=coeff)

    measurements = MeasurementResult(
        rng.randint(low=0, high=1 + 1, size=(100, nqubits)).tolist()
    )
    assert np.isclose(
        pauli._expectation_from_measurements(measurements),
        coeff,
    )


def test_expectation_from_measurements_two_qubits():
    measurements = MeasurementResult([[0, 1] * 1_000])

    z0 = PauliString(spec="Z", support=(0,))
    assert np.isclose(
        z0._expectation_from_measurements(measurements),
        1.0,
    )
    zi = PauliString(spec="ZI")
    assert np.isclose(
        zi._expectation_from_measurements(measurements),
        1.0,
    )

    z1 = PauliString(spec="Z", support=(1,))
    assert np.isclose(
        z1._expectation_from_measurements(measurements),
        -1.0,
    )
    iz = PauliString(spec="IZ")
    assert np.isclose(
        iz._expectation_from_measurements(measurements),
        -1.0,
    )

    zz = PauliString(spec="ZZ")
    assert np.isclose(
        zz._expectation_from_measurements(measurements),
        -1.0,
    )


def test_pstringcollection():
    x = PauliString(spec="X")
    iz = PauliString(spec="IZ")
    xz = PauliString(spec="XZ")
    xzixx = PauliString(spec="XZIXX")
    pauli_collection = PauliStringCollection(x, iz, xz, xzixx)

    assert pauli_collection.elements == [x, iz, xz, xzixx]
    assert pauli_collection.elements_by_weight == {
        1: Counter((x, iz)),
        2: Counter((xz,)),
        4: Counter((xzixx,)),
    }
    assert pauli_collection.min_weight() == 1
    assert pauli_collection.max_weight() == 4
    assert pauli_collection.support() == {0, 1, 3, 4}
    assert len(pauli_collection) == 4


def test_pstring_collection_empty():
    pauli_collection = PauliStringCollection()

    assert pauli_collection.elements == []
    assert pauli_collection.elements_by_weight == {}
    assert pauli_collection.min_weight() == 0
    assert pauli_collection.support() == set()
    assert len(pauli_collection) == 0


def test_pstring_collection_str():
    x = PauliString(spec="X")
    iz = PauliString(spec="IZ")
    pcol = PauliStringCollection(x, iz)
    assert str(pcol) == "X(q(0)) + Z(q(1))"

    xz = PauliString(spec="XZ", coeff=-2.4)
    pcol.add(xz)
    assert str(pcol) == "X(q(0)) + Z(q(1)) + (-2.4+0j)*X(q(0))*Z(q(1))"


def test_pstring_collection_add():
    pcol = PauliStringCollection()

    a = PauliString(spec="ZZ")
    assert pcol.can_add(a)
    pcol.add(a)
    assert pcol.elements == [a]

    b = PauliString(spec="ZIXZ")
    assert pcol.can_add(b)
    pcol.add(b)
    assert pcol.elements == [a, b]

    assert pcol.can_add(a)
    pcol.add(a)
    assert pcol.elements == [a, a, b]

    c = PauliString(spec="YY")
    assert not pcol.can_add(c)
    with pytest.raises(ValueError, match="Cannot add PauliString"):
        pcol.add(c)


def test_pstring_collection_len():
    x = PauliString(spec="X", support=(0,))
    y = PauliString(spec="Y", support=(1,))
    z = PauliString(spec="Z", support=(2,))
    assert len(PauliStringCollection(x, y, z)) == 3
    assert len(PauliStringCollection(x, x, x)) == 3
    assert len(PauliStringCollection(x, y)) == 2
    assert len(PauliStringCollection(x)) == 1
    assert len(PauliStringCollection()) == 0


def test_pstring_collection_eq():
    x = PauliString(spec="X")
    z = PauliString(spec="IZ")
    xz = PauliString(spec="XZ")
    xzz = PauliString(spec="IIIXZZ")

    assert PauliStringCollection(x, xzz) == PauliStringCollection(xzz, x)
    assert PauliStringCollection(x, z) != PauliStringCollection(x, xz)
    assert PauliStringCollection(x, z, xz) == PauliStringCollection(z, xz, x)
    assert PauliStringCollection() == PauliStringCollection()


def test_pstringcollection_expectation_from_measurements():
    measurements = MeasurementResult([[0, 0], [0, 0], [0, 1], [0, 1]])
    pset = PauliStringCollection(
        PauliString(spec="ZI", coeff=-2.0), PauliString(spec="IZ", coeff=5.0)
    )
    assert np.isclose(pset._expectation_from_measurements(measurements), -2.0)

    measurements = MeasurementResult([[1, 0], [1, 0], [0, 1], [0, 1]])
    pset = PauliStringCollection(
        PauliString(spec="ZI", coeff=-2.0), PauliString(spec="IZ", coeff=5.0)
    )
    assert np.isclose(pset._expectation_from_measurements(measurements), 0.0)


def test_pstringcollection_expectation_from_measurements_qubit_indices():
    measurements = MeasurementResult(
        [[0, 0], [0, 0], [0, 1], [0, 1]], qubit_indices=(1, 5)
    )
    pset = PauliStringCollection(
        PauliString(spec="Z", coeff=-2.0, support=(1,))
    )
    assert np.isclose(pset._expectation_from_measurements(measurements), -2.0)

    pset = PauliStringCollection(
        PauliString(spec="Z", coeff=-2.0, support=(5,))
    )
    assert np.isclose(pset._expectation_from_measurements(measurements), 0.0)


def test_spec():
    assert PauliString(spec="XIZYII").spec == "XZY"


def test_with_coeff():
    assert PauliString(spec="X").with_coeff(2).coeff == 2



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import cirq
import numpy as np
import pyquil
import pytest
import qiskit
from cirq import Circuit

from mitiq.pec.types import NoisyBasis, NoisyOperation, OperationRepresentation
from mitiq.utils import _equal

icirq = Circuit(cirq.I(cirq.LineQubit(0)))
xcirq = Circuit(cirq.X(cirq.LineQubit(0)))
ycirq = Circuit(cirq.Y(cirq.LineQubit(0)))
zcirq = Circuit(cirq.Z(cirq.LineQubit(0)))
hcirq = Circuit(cirq.H(cirq.LineQubit(0)))
cnotcirq = Circuit(cirq.CNOT(cirq.LineQubit(0), cirq.LineQubit(1)))


def test_init_with_cirq_circuit():
    real = np.zeros(shape=(4, 4))
    noisy_op = NoisyOperation(zcirq, real)
    assert isinstance(noisy_op._circuit, cirq.Circuit)

    assert noisy_op.qubits == (cirq.LineQubit(0),)
    assert np.allclose(noisy_op.channel_matrix, real)
    assert noisy_op.channel_matrix is not real

    assert noisy_op._native_type == "cirq"
    assert _equal(zcirq, noisy_op.circuit)
    assert _equal(noisy_op._native_circuit, noisy_op.circuit)


@pytest.mark.parametrize(
    "qubit",
    (cirq.LineQubit(0), cirq.GridQubit(1, 2), cirq.NamedQubit("Qubit")),
)
def test_init_with_different_qubits(qubit):
    ideal_op = Circuit(cirq.H.on(qubit))
    real = np.zeros(shape=(4, 4))
    noisy_op = NoisyOperation(ideal_op, real)

    assert isinstance(noisy_op._circuit, cirq.Circuit)
    assert _equal(
        noisy_op.circuit,
        cirq.Circuit(ideal_op),
        require_qubit_equality=True,
    )
    assert noisy_op.qubits == (qubit,)
    assert np.allclose(noisy_op.channel_matrix, real)
    assert noisy_op.channel_matrix is not real

    assert noisy_op._native_type == "cirq"
    assert _equal(cirq.Circuit(ideal_op), noisy_op.circuit)
    assert _equal(noisy_op._native_circuit, noisy_op.circuit)


def test_init_with_cirq_input():
    qreg = cirq.LineQubit.range(2)
    circ = cirq.Circuit(cirq.H.on(qreg[0]), cirq.CNOT.on(*qreg))
    real = np.zeros(shape=(16, 16))
    noisy_op = NoisyOperation(circ, real)

    assert isinstance(noisy_op._circuit, cirq.Circuit)
    assert _equal(noisy_op.circuit, circ, require_qubit_equality=True)
    assert set(noisy_op.qubits) == set(qreg)
    assert np.allclose(noisy_op.channel_matrix, real)
    assert noisy_op.channel_matrix is not real


def test_init_with_qiskit_circuit():
    qreg = qiskit.QuantumRegister(2)
    circ = qiskit.QuantumCircuit(qreg)
    _ = circ.h(qreg[0])
    _ = circ.cx(*qreg)

    cirq_qreg = cirq.LineQubit.range(2)
    cirq_circ = cirq.Circuit(cirq.H.on(cirq_qreg[0]), cirq.CNOT.on(*cirq_qreg))

    real = np.zeros(shape=(16, 16))
    noisy_op = NoisyOperation(circ, real)
    assert isinstance(noisy_op._circuit, cirq.Circuit)
    assert _equal(noisy_op._circuit, cirq_circ)
    assert _equal(noisy_op.circuit, cirq_circ)

    assert noisy_op.native_circuit == circ
    assert noisy_op._native_circuit == circ
    assert noisy_op._native_type == "qiskit"

    assert np.allclose(noisy_op.channel_matrix, real)
    assert noisy_op.channel_matrix is not real


@pytest.mark.parametrize(
    "gate",
    (
        cirq.H,
        cirq.H(cirq.LineQubit(0)),
        qiskit.circuit.library.HGate,
        qiskit.circuit.library.CHGate,
        pyquil.gates.H,
    ),
)
def test_init_with_gates_raises_error(gate):
    rng = np.random.RandomState(seed=1)
    with pytest.raises(TypeError, match="Failed to convert to an internal"):
        NoisyOperation(circuit=gate, channel_matrix=rng.rand(4, 4))


def test_init_with_pyquil_program():
    circ = pyquil.Program(pyquil.gates.H(0), pyquil.gates.CNOT(0, 1))

    cirq_qreg = cirq.LineQubit.range(2)
    cirq_circ = cirq.Circuit(cirq.H.on(cirq_qreg[0]), cirq.CNOT.on(*cirq_qreg))

    real = np.zeros(shape=(16, 16))
    noisy_op = NoisyOperation(circ, real)
    assert isinstance(noisy_op._circuit, cirq.Circuit)
    assert _equal(noisy_op._circuit, cirq_circ)
    assert _equal(noisy_op.circuit, cirq_circ)

    assert noisy_op.native_circuit == circ
    assert noisy_op._native_circuit == circ
    assert noisy_op._native_type == "pyquil"

    assert np.allclose(noisy_op.channel_matrix, real)
    assert noisy_op.channel_matrix is not real


def test_init_dimension_mismatch_error():
    ideal = cirq.Circuit(cirq.H.on(cirq.LineQubit(0)))
    real = np.zeros(shape=(3, 3))
    with pytest.raises(ValueError, match="has shape"):
        NoisyOperation(ideal, real)


def test_unknown_channel_matrix():
    qreg = qiskit.QuantumRegister(2)
    circ = qiskit.QuantumCircuit(qreg)
    _ = circ.h(qreg[0])
    _ = circ.cx(*qreg)

    cirq_qreg = cirq.LineQubit.range(2)
    cirq_circ = cirq.Circuit(cirq.H.on(cirq_qreg[0]), cirq.CNOT.on(*cirq_qreg))

    noisy_op = NoisyOperation(circ)
    assert isinstance(noisy_op._circuit, cirq.Circuit)
    assert _equal(noisy_op._circuit, cirq_circ)
    assert _equal(noisy_op.circuit, cirq_circ)

    assert noisy_op.native_circuit == circ
    assert noisy_op._native_circuit == circ
    assert noisy_op._native_type == "qiskit"

    with pytest.raises(ValueError, match="The channel matrix is unknown."):
        _ = noisy_op.channel_matrix


def test_add_simple():
    circuit1 = cirq.Circuit([cirq.X.on(cirq.NamedQubit("Q"))])
    circuit2 = cirq.Circuit([cirq.Y.on(cirq.NamedQubit("Q"))])

    super_op1 = np.random.rand(4, 4)
    super_op2 = np.random.rand(4, 4)

    noisy_op1 = NoisyOperation(circuit1, super_op1)
    noisy_op2 = NoisyOperation(circuit2, super_op2)

    noisy_op = noisy_op1 + noisy_op2

    correct = cirq.Circuit(
        [cirq.X.on(cirq.NamedQubit("Q")), cirq.Y.on(cirq.NamedQubit("Q"))],
    )

    assert _equal(noisy_op._circuit, correct, require_qubit_equality=True)
    assert np.allclose(noisy_op.channel_matrix, super_op2 @ super_op1)


def test_add_pyquil_noisy_operations():
    ideal = pyquil.Program(pyquil.gates.X(0))
    real = np.random.rand(4, 4)

    noisy_op1 = NoisyOperation(ideal, real)
    noisy_op2 = NoisyOperation(ideal, real)

    noisy_op = noisy_op1 + noisy_op2

    correct = cirq.Circuit([cirq.X.on(cirq.NamedQubit("Q"))] * 2)

    assert _equal(noisy_op._circuit, correct, require_qubit_equality=False)
    assert np.allclose(noisy_op.channel_matrix, real @ real)


def test_add_qiskit_noisy_operations():
    qreg = qiskit.QuantumRegister(1)
    ideal = qiskit.QuantumCircuit(qreg)
    _ = ideal.x(qreg)
    real = np.random.rand(4, 4)

    noisy_op1 = NoisyOperation(ideal, real)
    noisy_op2 = NoisyOperation(ideal, real)

    noisy_op = noisy_op1 + noisy_op2

    correct = cirq.Circuit([cirq.X.on(cirq.NamedQubit("Q"))] * 2)

    assert _equal(noisy_op._circuit, correct, require_qubit_equality=False)
    assert np.allclose(noisy_op.channel_matrix, real @ real)


def test_add_bad_type():
    ideal = cirq.Circuit([cirq.X.on(cirq.NamedQubit("Q"))])
    real = np.random.rand(4, 4)

    noisy_op = NoisyOperation(ideal, real)

    with pytest.raises(ValueError, match="must be a NoisyOperation"):
        noisy_op + ideal


def test_add_noisy_operation_no_channel_matrix():
    noisy_op1 = NoisyOperation(cirq.Circuit([cirq.X.on(cirq.NamedQubit("Q"))]))
    noisy_op2 = NoisyOperation(
        cirq.Circuit([cirq.X.on(cirq.NamedQubit("Q"))]),
        channel_matrix=np.random.rand(4, 4),
    )

    with pytest.raises(ValueError):
        (noisy_op1 + noisy_op2).channel_matrix


def test_noisy_operation_str():
    noisy_op = NoisyOperation(circuit=icirq, channel_matrix=np.identity(4))
    assert isinstance(noisy_op.__str__(), str)


def test_noisy_basis_deprecation_error():
    with pytest.raises(NotImplementedError, match="NoisyBasis"):
        NoisyBasis()
    with pytest.raises(NotImplementedError, match="NoisyBasis"):
        NoisyBasis(zcirq, xcirq)


def get_test_representation():
    ideal = cirq.Circuit(cirq.H(cirq.LineQubit(0)))

    noisy_xop = NoisyOperation(
        circuit=xcirq, channel_matrix=np.zeros(shape=(4, 4))
    )
    noisy_zop = NoisyOperation(
        circuit=zcirq, channel_matrix=np.zeros(shape=(4, 4))
    )

    decomp = OperationRepresentation(
        ideal,
        [noisy_xop, noisy_zop],
        [0.5, -0.5],
    )
    return ideal, noisy_xop, noisy_zop, decomp


def test_representation_simple():
    ideal, noisy_xop, noisy_zop, decomp = get_test_representation()

    assert _equal(decomp.ideal, ideal)
    assert decomp.coeffs == [0.5, -0.5]
    assert np.allclose(decomp.distribution, np.array([0.5, 0.5]))
    assert np.isclose(decomp.norm, 1.0)
    assert isinstance(decomp.basis_expansion[0][0], float)
    assert set(decomp.noisy_operations) == {noisy_xop, noisy_zop}


def test_representation_bad_type():
    ideal = cirq.Circuit(cirq.H(cirq.LineQubit(0)))

    noisy_xop = NoisyOperation(
        circuit=xcirq, channel_matrix=np.zeros(shape=(4, 4))
    )

    with pytest.raises(TypeError, match="All elements of `noisy_operations`"):
        OperationRepresentation(
            ideal=ideal,
            noisy_operations=[0.1],
            coeffs=[0.1],
        )
    with pytest.raises(TypeError, match="All elements of `coeffs` must"):
        OperationRepresentation(
            ideal=ideal,
            noisy_operations=[noisy_xop],
            coeffs=["x"],
        )


def test_representation_sample():
    _, noisy_xop, noisy_zop, decomp = get_test_representation()

    for _ in range(10):
        noisy_op, sign, coeff = decomp.sample()
        assert sign in (-1, 1)
        assert coeff in (-0.5, 0.5)
        assert noisy_op in (noisy_xop, noisy_zop)
        case_one = noisy_op == noisy_xop and coeff == 0.5
        case_two = noisy_op == noisy_zop and coeff == -0.5
        assert case_one or case_two


def test_representation_sample_seed():
    _, noisy_xop, noisy_zop, decomp = get_test_representation()

    seed1 = np.random.RandomState(seed=1)
    seed2 = np.random.RandomState(seed=1)
    for _ in range(10):
        _, sign1, coeff1 = decomp.sample(random_state=seed1)
        _, sign2, coeff2 = decomp.sample(random_state=seed2)

        assert sign1 == sign2
        assert np.isclose(coeff1, coeff2)


def test_representation_sample_bad_seed_type():
    _, _, _, decomp = get_test_representation()
    with pytest.raises(TypeError, match="should be of type"):
        decomp.sample(random_state="8")


def test_representation_sample_zero_coefficient():
    ideal = cirq.Circuit(cirq.H(cirq.LineQubit(0)))

    noisy_xop = NoisyOperation(
        circuit=xcirq, channel_matrix=np.zeros(shape=(4, 4))
    )
    noisy_zop = NoisyOperation(
        circuit=zcirq, channel_matrix=np.zeros(shape=(4, 4))
    )

    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.5, 0.0],  # 0 term should never be sampled.
    )

    random_state = np.random.RandomState(seed=1)
    for _ in range(500):
        noisy_op, sign, coeff = decomp.sample(random_state=random_state)
        assert sign == 1
        assert coeff == 0.5
        assert np.allclose(
            cirq.unitary(noisy_op.circuit),
            cirq.unitary(cirq.X),
        )


def test_print_cirq_operation_representation():
    ideal = cirq.Circuit(cirq.H(cirq.LineQubit(0)))

    noisy_xop = NoisyOperation(
        circuit=xcirq, channel_matrix=np.zeros(shape=(4, 4))
    )
    noisy_zop = NoisyOperation(
        circuit=zcirq, channel_matrix=np.zeros(shape=(4, 4))
    )
    # Positive first coefficient
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.5, 0.5],
    )
    expected = r"0: ───H─── = 0.500*(0: ───X───)+0.500*(0: ───Z───)"
    assert str(decomp) == expected
    # Negative first coefficient
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[-0.5, 1.5],
    )
    expected = r"0: ───H─── = -0.500*(0: ───X───)+1.500*(0: ───Z───)"
    assert str(decomp) == expected
    # Empty representation
    decomp = OperationRepresentation(ideal, [], [])
    expected = r"0: ───H─── = 0.000"
    assert str(decomp) == expected
    # Small coefficient approximation
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[1.00001, 0.00001],
    )
    expected = r"0: ───H─── = 1.000*(0: ───X───)"
    assert str(decomp) == expected
    # Small coefficient approximation different position
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop, noisy_zop],
        coeffs=[0.00001, 1.00001],
    )
    expected = r"0: ───H─── = 1.000*(0: ───Z───)"
    # Small coefficient approximation different position
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop],
        coeffs=[0.00001],
    )
    expected = r"0: ───H─── = 0.000"
    assert str(decomp) == expected


def test_print_operation_representation_two_qubits():
    qreg = cirq.LineQubit.range(2)
    ideal = cirq.Circuit(cirq.CNOT(*qreg))

    noisy_a = NoisyOperation(
        circuit=cirq.Circuit(
            cirq.H.on_each(qreg), cirq.CNOT(*qreg), cirq.H.on_each(qreg)
        )
    )
    noisy_b = NoisyOperation(
        circuit=cirq.Circuit(
            cirq.Z.on_each(qreg),
            cirq.CNOT(*qreg),
            cirq.Z.on_each(qreg),
        )
    )
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_a, noisy_b],
        coeffs=[0.5, 0.5],
    )
    expected = f"""
0: ───@───
      │
1: ───X─── ={" "}

0.500
0: ───H───@───H───
          │
1: ───H───X───H───

+0.500
0: ───Z───@───Z───
          │
1: ───Z───X───Z───"""
    # Remove initial newline
    expected = expected[1:]
    assert str(decomp) == expected


def test_print_operation_representation_two_qubits_neg():
    qreg = cirq.LineQubit.range(2)
    ideal = cirq.Circuit(cirq.CNOT(*qreg))

    noisy_a = NoisyOperation(
        circuit=cirq.Circuit(
            cirq.H.on_each(qreg[0]), cirq.CNOT(*qreg), cirq.H.on_each(qreg[1])
        )
    )
    noisy_b = NoisyOperation(circuit=cirq.Circuit(cirq.Z.on_each(*qreg)))
    decomp = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_a, noisy_b],
        coeffs=[-0.5, 1.5],
    )
    expected = f"""
0: ───@───
      │
1: ───X─── ={" "}

-0.500
0: ───H───@───────
          │
1: ───────X───H───

+1.500
0: ───Z───

1: ───Z───"""

    # Remove initial newline
    expected = expected[1:]
    assert str(decomp) == expected


def test_equal_method_of_representations():
    q = cirq.LineQubit(0)
    ideal = cirq.Circuit(cirq.H(q))
    noisy_xop_a = NoisyOperation(
        circuit=cirq.Circuit(cirq.X(q)),
        channel_matrix=np.zeros(shape=(4, 4)),
    )
    noisy_zop_a = NoisyOperation(
        circuit=cirq.Circuit(cirq.Z(q)),
        channel_matrix=np.zeros(shape=(4, 4)),
    )
    rep_a = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop_a, noisy_zop_a],
        coeffs=[0.5, 0.5],
    )
    noisy_xop_b = NoisyOperation(
        circuit=cirq.Circuit(cirq.X(q)),
        channel_matrix=np.ones(shape=(4, 4)),
    )
    noisy_zop_b = NoisyOperation(
        circuit=cirq.Circuit(cirq.Z(q)),
        channel_matrix=np.ones(shape=(4, 4)),
    )
    rep_b = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop_b, noisy_zop_b],
        coeffs=[0.5, 0.5],
    )
    # Equal representation up to real superoperators
    assert rep_a == rep_b
    # Different ideal
    ideal_b = cirq.Circuit(cirq.X(q))
    rep_b = OperationRepresentation(
        ideal=ideal_b,
        noisy_operations=[noisy_xop_b, noisy_zop_b],
        coeffs=[0.5, 0.5],
    )
    assert rep_a != rep_b
    # Different type
    q_b = qiskit.QuantumRegister(1)
    ideal_b = qiskit.QuantumCircuit(q_b)
    ideal_b.x(q_b)
    noisy_opx = NoisyOperation(ideal_b)
    rep_b = OperationRepresentation(
        ideal=ideal_b,
        noisy_operations=[noisy_opx, noisy_opx],
        coeffs=[0.5, 0.5],
    )
    assert rep_a != rep_b
    # Different length
    rep_b = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop_b],
        coeffs=[0.5],
    )
    assert rep_a != rep_b
    # Different operations
    noisy_diff = NoisyOperation(circuit=cirq.Circuit(cirq.H(q)))
    rep_b = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop_b, noisy_diff],
        coeffs=[0.5, 0.5],
    )
    assert rep_a != rep_b
    # Different coefficients
    rep_b = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop_b, noisy_zop_b],
        coeffs=[0.7, 0.5],
    )
    assert rep_a != rep_b
    # Different value of is_qubit_dependent
    rep_b = OperationRepresentation(
        ideal=ideal,
        noisy_operations=[noisy_xop_a, noisy_zop_a],
        coeffs=[0.5, 0.5],
        is_qubit_dependent=False,
    )
    assert rep_a != rep_b


def test_operation_representation_warnings():
    with pytest.warns(UserWarning, match="different from 1"):
        OperationRepresentation(
            ideal=xcirq,
            noisy_operations=[NoisyOperation(xcirq), NoisyOperation(zcirq)],
            coeffs=[0.5, 0.1],
        )


def test_different_qubits_error():
    """Ideal operation and noisy operations must have equal qubits."""

    with pytest.raises(ValueError, match="must act on the same qubits"):
        OperationRepresentation(
            ideal=cirq.Circuit(cirq.X(cirq.NamedQubit("a"))),
            noisy_operations=[NoisyOperation(xcirq), NoisyOperation(zcirq)],
            coeffs=[0.5, 0.5],
        )


def test_different_length_error():
    """The number of coefficients must be equal to the number of noisy
    operations.
    """
    with pytest.raises(ValueError, match="must have equal length"):
        OperationRepresentation(
            ideal=cirq.Circuit(cirq.X(cirq.LineQubit(0))),
            noisy_operations=[NoisyOperation(xcirq), NoisyOperation(zcirq)],
            coeffs=[0.5, 0.5, 0.4],
        )



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for RAW (no error mitigation)."""

import functools

import cirq
import numpy as np

from mitiq import Executor, Observable, PauliString, raw
from mitiq.interface import mitiq_cirq


def test_raw():
    circuit = cirq.Circuit(cirq.H.on(cirq.LineQubit(0)))
    observable = Observable(PauliString("X"))
    executor = Executor(mitiq_cirq.compute_density_matrix)
    raw_value = raw.execute(circuit, executor, observable)

    assert isinstance(raw_value, float)
    assert executor.executed_circuits == [circuit]

    compute_density_matrix_noiseless = functools.partial(
        mitiq_cirq.compute_density_matrix, noise_level=(0.0,)
    )
    executor_noiseless = Executor(compute_density_matrix_noiseless)
    true_value = raw.execute(circuit, executor_noiseless, observable)
    assert np.isclose(true_value, 1.0)
    assert executor_noiseless.executed_circuits == [circuit]



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for virtual distillation."""

from unittest.mock import MagicMock

import cirq

from mitiq import MeasurementResult
from mitiq.vd import combine_results, construct_circuits, execute_with_vd


def test_construct_circuits_adds_intended_qubits_and_measurements():
    qubits = cirq.LineQubit.range(6)
    test_circuit = cirq.Circuit(
        cirq.X(qubits[0]),
        cirq.Y(qubits[1]),
        cirq.Z(qubits[2]),
    )
    vd_circuit = construct_circuits(test_circuit)

    assert vd_circuit.all_qubits() == set(qubits)
    assert len(vd_circuit.all_measurement_key_objs()) == 1


def test_combine_results_gives_correct_number_of_expvals():
    mr = MeasurementResult.from_counts(
        {"0000": 124, "1111": 103, "0001": 89, "1110": 104}
    )
    result = combine_results(mr)

    assert len(result) == 2


def test_execute_with_vd_makes_single_executor_call():
    qubits = cirq.LineQubit.range(6)
    test_circuit = cirq.Circuit(
        cirq.X(qubits[0]),
        cirq.Y(qubits[1]),
        cirq.Z(qubits[2]),
    )

    mock_executor = MagicMock(return_value=MeasurementResult(["000", "001"]))

    _ = execute_with_vd(test_circuit, mock_executor)

    assert mock_executor.call_count == 1



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tests for mirror quantum volume circuits."""

import cirq
import pytest

from mitiq import SUPPORTED_PROGRAM_TYPES
from mitiq.benchmarks.mirror_qv_circuits import generate_mirror_qv_circuit


@pytest.mark.parametrize(
    "depth_num",
    [1, 2, 3, 4, 5, 6, 7, 8],
)
def test_generate_mirror_qv_circuit(depth_num):
    """Check the circuit output."""
    test_circ = generate_mirror_qv_circuit(4, depth_num)

    # check bitstring is all 0's
    bit_test = cirq.Simulator().run(
        test_circ + cirq.measure(test_circ.all_qubits()), repetitions=1000
    )
    test_bitstring = list(bit_test.measurements.values())[0][0].tolist()
    expected_bitstring = [0] * 4
    assert test_bitstring == expected_bitstring


def test_bad_depth_number():
    """Check if an unacceptable depth number rasies an error."""
    for n in (-1, 0):
        with pytest.raises(
            ValueError, match="{} is invalid for the generated circuit depth."
        ):
            generate_mirror_qv_circuit(3, n)


@pytest.mark.parametrize("return_type", SUPPORTED_PROGRAM_TYPES.keys())
def test_volume_conversion(return_type):
    """Check generated circuit's return type."""
    circuit = generate_mirror_qv_circuit(4, 3, return_type=return_type)
    assert return_type in circuit.__module__


def test_generate_model_circuit_with_seed():
    """Test that a model circuit is determined by its seed."""
    circuit_1 = generate_mirror_qv_circuit(4, 3, seed=1)
    circuit_2 = generate_mirror_qv_circuit(4, 3, seed=1)
    circuit_3 = generate_mirror_qv_circuit(4, 3, seed=2)

    assert cirq.approx_eq(circuit_1, circuit_2, atol=1e-12)
    assert circuit_2 != circuit_3


def test_circuit_decomposition():
    """Test that decomposed circuit consists of gates in default cirq gatest.
    As defined in cirq.protocols.decompose_protocol, this default gateset is
        ops.XPowGate,
        ops.YPowGate,
        ops.ZPowGate,
        ops.CZPowGate,
        ops.MeasurementGate,
        ops.GlobalPhaseGate
    """
    circuit = generate_mirror_qv_circuit(4, 3, decompose=True)
    for op in [operation for moment in circuit for operation in moment]:
        assert op in cirq.protocols.decompose_protocol.DECOMPOSE_TARGET_GATESET



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest
from cirq import (
    CCNOT,
    CNOT,
    CZ,
    ISWAP,
    SWAP,
    Circuit,
    DepolarizingChannel,
    Gate,
    H,
    LineQubit,
    MeasurementGate,
    X,
    Y,
    Z,
)

from mitiq.interface import convert_from_mitiq, convert_to_mitiq
from mitiq.pec.channels import _circuit_to_choi, _operation_to_choi
from mitiq.pec.representations import (
    represent_operation_with_global_depolarizing_noise,
    represent_operation_with_local_depolarizing_noise,
    represent_operations_in_circuit_with_global_depolarizing_noise,
    represent_operations_in_circuit_with_local_depolarizing_noise,
)
from mitiq.pec.representations.depolarizing import (
    global_depolarizing_kraus,
    local_depolarizing_kraus,
)
from mitiq.utils import _equal


def single_qubit_depolarizing_overhead(noise_level: float) -> float:
    """See :cite:`Temme_2017_PRL` for more information.

    Args:
        noise_level: multiplier of noise level in :cite:`Temme_2017_PRL`

    Returns:
        Depolarizing overhead value with noise level considered.
    """
    epsilon = 4 / 3 * noise_level
    return (1 + epsilon / 2) / (1 - epsilon)


def two_qubit_depolarizing_overhead(noise_level: float) -> float:
    """See :cite:`Temme_2017_PRL` for more information.

    Args:
        noise_level: multiplier of noise level in :cite:`Temme_2017_PRL`

    Returns:
        Depolarizing overhead value with noise level considered.
    """
    epsilon = 16 / 15 * noise_level
    return (1 + 7 * epsilon / 8) / (1 - epsilon)


@pytest.mark.parametrize("noise", [0, 0.1, 0.7])
@pytest.mark.parametrize("gate", [X, Y, Z, H])
def test_single_qubit_representation_norm(gate: Gate, noise: float):
    q = LineQubit(0)
    optimal_norm = single_qubit_depolarizing_overhead(noise)
    norm = represent_operation_with_global_depolarizing_noise(
        Circuit(gate(q)),
        noise,
    ).norm
    assert np.isclose(optimal_norm, norm)


@pytest.mark.parametrize("noise", [0, 0.1, 0.7])
@pytest.mark.parametrize("gate", (CZ, CNOT, ISWAP, SWAP))
def test_two_qubit_representation_norm(gate: Gate, noise: float):
    qreg = LineQubit.range(2)
    optimal_norm = two_qubit_depolarizing_overhead(noise)
    norm = represent_operation_with_global_depolarizing_noise(
        Circuit(gate(*qreg)),
        noise,
    ).norm
    assert np.isclose(optimal_norm, norm)


def test_three_qubit_depolarizing_representation_error():
    q0, q1, q2 = LineQubit.range(3)
    with pytest.raises(ValueError):
        represent_operation_with_global_depolarizing_noise(
            Circuit(CCNOT(q0, q1, q2)),
            0.05,
        )


@pytest.mark.parametrize("noise", [0, 0.1, 0.7])
@pytest.mark.parametrize("gate", [X, Y, Z, H, CZ, CNOT, ISWAP, SWAP])
def test_depolarizing_representation_with_choi(gate: Gate, noise: float):
    """Tests the representation by comparing exact Choi matrices."""
    qreg = LineQubit.range(gate.num_qubits())
    ideal_choi = _operation_to_choi(gate.on(*qreg))
    op_rep = represent_operation_with_global_depolarizing_noise(
        Circuit(gate.on(*qreg)),
        noise,
    )
    choi_components = []
    for coeff, noisy_op in op_rep.basis_expansion:
        implementable_circ = noisy_op.circuit
        # Apply noise after each sequence.
        # NOTE: noise is not applied after each operation.
        depolarizing_op = DepolarizingChannel(noise, len(qreg))(*qreg)
        implementable_circ.append(depolarizing_op)
        sequence_choi = _circuit_to_choi(implementable_circ)
        choi_components.append(coeff * sequence_choi)
    combination_choi = np.sum(choi_components, axis=0)
    assert np.allclose(ideal_choi, combination_choi, atol=10**-6)


@pytest.mark.parametrize("noise", [0, 0.1, 0.7])
@pytest.mark.parametrize("gate", [X, Y, Z, H, CZ, CNOT, ISWAP, SWAP])
def test_local_depolarizing_representation_with_choi(gate: Gate, noise: float):
    """Tests the representation by comparing exact Choi matrices."""
    qreg = LineQubit.range(gate.num_qubits())
    ideal_choi = _operation_to_choi(gate.on(*qreg))
    op_rep = represent_operation_with_local_depolarizing_noise(
        Circuit(gate.on(*qreg)),
        noise,
    )
    choi_components = []
    for coeff, noisy_op in op_rep.basis_expansion:
        implementable_circ = noisy_op.circuit
        # The representation assume local noise on each qubit.
        depolarizing_op = DepolarizingChannel(noise).on_each(*qreg)
        # Apply noise after each sequence.
        # NOTE: noise is not applied after each operation.
        implementable_circ.append(depolarizing_op)
        sequence_choi = _circuit_to_choi(implementable_circ)
        choi_components.append(coeff * sequence_choi)
    combination_choi = np.sum(choi_components, axis=0)
    assert np.allclose(ideal_choi, combination_choi, atol=10**-6)


def test_three_qubit_local_depolarizing_representation_error():
    q0, q1, q2 = LineQubit.range(3)
    with pytest.raises(ValueError):
        represent_operation_with_local_depolarizing_noise(
            Circuit(CCNOT(q0, q1, q2)),
            0.05,
        )


@pytest.mark.parametrize("circuit_type", ["cirq", "qiskit", "pyquil"])
def test_represent_operations_in_circuit_global(circuit_type: str):
    """Tests all operation representations are created."""
    qreg = LineQubit.range(2)
    circ_mitiq = Circuit([CNOT(*qreg), H(qreg[0]), Y(qreg[1]), CNOT(*qreg)])
    circ = convert_from_mitiq(circ_mitiq, circuit_type)

    reps = represent_operations_in_circuit_with_global_depolarizing_noise(
        ideal_circuit=circ,
        noise_level=0.1,
    )

    # For each operation in circ we should find its representation
    for op in convert_to_mitiq(circ)[0].all_operations():
        found = False
        for rep in reps:
            if _equal(rep.ideal, Circuit(op), require_qubit_equality=True):
                found = True
        assert found

    # The number of reps. should match the number of unique operations
    assert len(reps) == 3


@pytest.mark.parametrize("circuit_type", ["cirq", "qiskit", "pyquil"])
def test_represent_operations_in_circuit_local(circuit_type: str):
    """Tests all operation representations are created."""
    qreg = LineQubit.range(2)
    circ_mitiq = Circuit([CNOT(*qreg), H(qreg[0]), Y(qreg[1]), CNOT(*qreg)])
    circ = convert_from_mitiq(circ_mitiq, circuit_type)

    reps = represent_operations_in_circuit_with_local_depolarizing_noise(
        ideal_circuit=circ,
        noise_level=0.1,
    )

    for op in convert_to_mitiq(circ)[0].all_operations():
        found = False
        for rep in reps:
            if _equal(rep.ideal, Circuit(op), require_qubit_equality=True):
                found = True
        assert found

    # The number of reps. should match the number of unique operations
    assert len(reps) == 3


@pytest.mark.parametrize(
    "rep_function",
    [
        represent_operations_in_circuit_with_local_depolarizing_noise,
        represent_operations_in_circuit_with_global_depolarizing_noise,
    ],
)
@pytest.mark.parametrize("circuit_type", ["cirq", "qiskit", "pyquil"])
def test_represent_operations_in_circuit_with_measurements(
    circuit_type: str,
    rep_function,
):
    """Tests measurements in circuit are ignored (not represented)."""
    q0, q1 = LineQubit.range(2)
    circ_mitiq = Circuit(
        X(q1),
        MeasurementGate(num_qubits=1)(q0),
        X(q1),
        MeasurementGate(num_qubits=1)(q0),
    )
    circ = convert_from_mitiq(circ_mitiq, circuit_type)

    reps = rep_function(ideal_circuit=circ, noise_level=0.1)

    for op in convert_to_mitiq(circ)[0].all_operations():
        found = False
        for rep in reps:
            if _equal(rep.ideal, Circuit(op), require_qubit_equality=True):
                found = True
        if isinstance(op.gate, MeasurementGate):
            assert not found
        else:
            assert found

    # Number of unique gates excluding measurement gates
    assert len(reps) == 1


@pytest.mark.parametrize(
    "kraus_function",
    [global_depolarizing_kraus, local_depolarizing_kraus],
)
def test_depolarizing_kraus(kraus_function):
    expected = [
        [[0.5, 0.0], [0.0, 0.5]],
        [[0.0 + 0.0j, 0.5 + 0.0j], [0.5 + 0.0j, 0.0 + 0.0j]],
        [[0.0 + 0.0j, 0.0 - 0.5j], [0.0 + 0.5j, 0.0 + 0.0j]],
        [[0.5 + 0.0j, 0.0 + 0.0j], [0.0 + 0.0j, -0.5 + 0.0j]],
    ]
    assert np.allclose(kraus_function(3 / 4, 1), expected)
    # Test normalization of kraus operators
    for num_qubits in (1, 2, 3):
        for noise_level in (0.1, 1):
            kraus_ops = kraus_function(noise_level, num_qubits)
            dual_channel = sum([k.conj().T @ k for k in kraus_ops])
            assert np.allclose(dual_channel, np.eye(2**num_qubits))



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for Pennylane <-> Cirq conversions."""

import cirq
import numpy as np
import pennylane as qml
import pytest

from mitiq.interface.mitiq_pennylane import (
    UnsupportedQuantumTapeError,
    from_pennylane,
    to_pennylane,
)
from mitiq.utils import _equal


def test_from_pennylane():
    with qml.tape.QuantumTape() as tape:
        qml.CNOT(wires=[0, 1])

    circuit = from_pennylane(tape)
    correct = cirq.Circuit(cirq.CNOT(*cirq.LineQubit.range(2)))

    assert _equal(circuit, correct, require_qubit_equality=False)


def test_from_pennylane_unsupported_tapes():
    with qml.tape.QuantumTape() as tape:
        qml.CZ(wires=[0, "a"])

    with pytest.raises(UnsupportedQuantumTapeError, match="could not sort"):
        from_pennylane(tape)


def test_no_variance():
    with qml.tape.QuantumTape() as tape:
        qml.CNOT(wires=[0, 1])
        qml.expval(qml.PauliZ(0))

    with pytest.raises(
        UnsupportedQuantumTapeError,
        match="Measurements are not supported on the input tape.",
    ):
        from_pennylane(tape)


@pytest.mark.parametrize("random_state", range(10))
def test_to_from_pennylane(random_state):
    circuit = cirq.testing.random_circuit(
        qubits=4, n_moments=2, op_density=1, random_state=random_state
    )

    converted = from_pennylane(to_pennylane(circuit))
    # Gates (e.g. iSWAP) aren't guaranteed to be preserved. Check unitary
    # instead of circuit equality.
    cirq.testing.assert_allclose_up_to_global_phase(
        cirq.unitary(converted), cirq.unitary(circuit), atol=1e-7
    )


def test_to_from_pennylane_cnot_same_gates():
    qreg = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(cirq.CNOT(*qreg))
    converted = from_pennylane(to_pennylane(circuit))
    assert _equal(circuit, converted, require_qubit_equality=False)


def test_to_from_pennylane_identity():
    q = cirq.LineQubit(0)
    # Empty circuit
    circuit = cirq.Circuit()
    converted = from_pennylane(to_pennylane(circuit))
    assert _equal(circuit, converted, require_qubit_equality=False)
    circuit = cirq.Circuit(cirq.I(q))
    # Identity gate
    converted = from_pennylane(to_pennylane(circuit))
    # TODO: test circuit equality after Identity operation will be added
    # to PennyLane (https://github.com/PennyLaneAI/pennylane/issues/1632)
    assert np.allclose(cirq.unitary(circuit), cirq.unitary(converted))


def test_non_consecutive_wires_error():
    with qml.tape.QuantumTape() as tape:
        qml.CNOT(wires=[0, 2])
    with pytest.raises(
        UnsupportedQuantumTapeError,
        match="contiguously pack",
    ):
        from_pennylane(tape)


def test_integration():
    gates = [
        qml.PauliX(wires=0),
        qml.PauliY(wires=0),
        qml.PauliZ(wires=0),
        qml.S(wires=0),
        qml.T(wires=0),
        qml.RX(0.4, wires=0),
        qml.RY(0.4, wires=0),
        qml.RZ(0.4, wires=0),
        qml.Hadamard(wires=0),
        qml.Rot(0.4, 0.5, 0.6, wires=1),
        qml.CRot(0.4, 0.5, 0.6, wires=(0, 1)),
        qml.Toffoli(wires=(0, 1, 2)),
        qml.SWAP(wires=(0, 1)),
        qml.CSWAP(wires=(0, 1, 2)),
        qml.U1(0.4, wires=0),
        qml.U2(0.4, 0.5, wires=0),
        qml.U3(0.4, 0.5, 0.6, wires=0),
        qml.CRX(0.4, wires=(0, 1)),
        qml.CRY(0.4, wires=(0, 1)),
        qml.CRZ(0.4, wires=(0, 1)),
    ]

    layers = 3
    np.random.seed(1967)
    gates_per_layers = [np.random.permutation(gates) for _ in range(layers)]

    with qml.tape.QuantumTape() as tape:
        np.random.seed(1967)
        for gates in gates_per_layers:
            for gate in gates:
                qml.apply(gate)

    base_circ = from_pennylane(tape)
    tape_recovered = to_pennylane(base_circ)
    circ_recovered = from_pennylane(tape_recovered)
    u_1 = cirq.unitary(base_circ)
    u_2 = cirq.unitary(circ_recovered)
    cirq.testing.assert_allclose_up_to_global_phase(u_1, u_2, atol=0)



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import os

from mitiq.typing import SUPPORTED_PROGRAM_TYPES


def test_supported_program_types_definition():
    directory_of_this_file = os.path.dirname(os.path.abspath(__file__))
    with open(f"{directory_of_this_file}/../../INTEGRATIONS.txt", "r") as file:
        integrations_from_setup = file.read().splitlines()

    assert list(SUPPORTED_PROGRAM_TYPES.keys()) == integrations_from_setup



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for converting to/from Mitiq's internal circuit representation."""

from collections.abc import Callable, Collection
from functools import wraps
from typing import (
    Any,
    Concatenate,
    ParamSpec,
    TypeVar,
    cast,
)

import cirq

from mitiq import QPROGRAM, SUPPORTED_PROGRAM_TYPES


class UnsupportedCircuitError(Exception):
    pass


class CircuitConversionError(Exception):
    pass


FROM_MITIQ_DICT: dict[str, Callable[[cirq.Circuit], Any]]
try:
    FROM_MITIQ_DICT
except NameError:
    FROM_MITIQ_DICT = {}

TO_MITIQ_DICT: dict[str, Callable[[Any], cirq.Circuit]]
try:
    TO_MITIQ_DICT
except NameError:
    TO_MITIQ_DICT = {}


def register_mitiq_converters(
    package_name: str,
    *,
    convert_to_function: Callable[[cirq.Circuit], Any],
    convert_from_function: Callable[[Any], cirq.Circuit],
) -> None:
    """Registers converters for unsupported circuit types.

    Args:
        package_name: A quantum circuit module name that is not currently
            supported by Mitiq. Note: this name should be the same as the
            return from "circuit".__module__.
                 See mitiq.SUPPORTED_PROGRAM_TYPES.
        convert_to_function: User specified function to convert to an
            unsupported circuit type. This function returns a Non-Mitiq
            circuit.
        convert_function: User specified function to convert from an
            unsupported circuit type. This function returns a Mitiq/Cirq
            circuit.
    """
    FROM_MITIQ_DICT[package_name] = convert_to_function
    TO_MITIQ_DICT[package_name] = convert_from_function


def convert_to_mitiq(circuit: QPROGRAM) -> tuple[cirq.Circuit, str]:
    """Converts any valid input circuit to a mitiq circuit.

    Args:
        circuit: Any quantum circuit object supported by mitiq.
                 See mitiq.SUPPORTED_PROGRAM_TYPES.

    Raises:
        UnsupportedCircuitError: If the input circuit is not supported.

    Returns:
        circuit: Mitiq circuit equivalent to input circuit.
        input_circuit_type: Type of input circuit represented by a string.
    """
    conversion_function: Callable[[Any], cirq.Circuit]

    try:
        package = circuit.__module__
    except AttributeError:
        raise UnsupportedCircuitError(
            "Could not determine the package of the input circuit."
        )

    if "qiskit" in package:
        from mitiq.interface.mitiq_qiskit.conversions import from_qiskit

        input_circuit_type = "qiskit"
        conversion_function = from_qiskit
    elif "pyquil" in package:
        from mitiq.interface.mitiq_pyquil.conversions import from_pyquil

        input_circuit_type = "pyquil"
        conversion_function = from_pyquil
    elif "braket" in package:
        from mitiq.interface.mitiq_braket.conversions import from_braket

        input_circuit_type = "braket"
        conversion_function = from_braket
    elif "pennylane" in package:
        from mitiq.interface.mitiq_pennylane.conversions import from_pennylane

        input_circuit_type = "pennylane"
        conversion_function = from_pennylane
    elif "qibo" in package:
        from mitiq.interface.mitiq_qibo.conversions import from_qibo

        input_circuit_type = "qibo"
        conversion_function = from_qibo

    elif package in TO_MITIQ_DICT:
        input_circuit_type = package
        conversion_function = TO_MITIQ_DICT[package]

    elif isinstance(circuit, cirq.Circuit):
        input_circuit_type = "cirq"

        def conversion_function(circ: cirq.Circuit) -> cirq.Circuit:
            return circ

    else:
        raise UnsupportedCircuitError(
            f"Circuit from module {package} is not supported.\n\n"
            f"Please register converters with register_mitiq_converters(),"
            f"\n or specify a supported Circuit type:"
            f"\n {SUPPORTED_PROGRAM_TYPES}"
        )

    try:
        mitiq_circuit = conversion_function(circuit)
    except Exception:
        raise CircuitConversionError(
            "Circuit could not be converted to an internal Mitiq circuit. "
            "This may be because the circuit contains custom gates or Pragmas "
            "(pyQuil). If you think this is a bug or that this circuit should "
            "be supported, you can open an issue at "
            "https://github.com/unitaryfoundation/mitiq. \n\n Provided "
            f"circuit has type {type(circuit)} and is:\n\n{circuit}\n\n "
            f"Circuit types supported by Mitiq are "
            "\n{SUPPORTED_PROGRAM_TYPES}."
        )

    return mitiq_circuit, input_circuit_type


def convert_from_mitiq(
    circuit: cirq.Circuit, conversion_type: str
) -> QPROGRAM:
    """Converts a Mitiq circuit to a type specified by the conversion type.

    Args:
        circuit: Mitiq circuit to convert.
        conversion_type: String specifier for the converted circuit type.
    """
    conversion_type = conversion_type.lower()
    conversion_function: Callable[[cirq.Circuit], QPROGRAM]
    if conversion_type == "qiskit":
        from mitiq.interface.mitiq_qiskit.conversions import to_qiskit

        conversion_function = to_qiskit
    elif conversion_type == "pyquil":
        from mitiq.interface.mitiq_pyquil.conversions import to_pyquil

        conversion_function = to_pyquil
    elif conversion_type == "braket":
        from mitiq.interface.mitiq_braket.conversions import to_braket

        conversion_function = to_braket
    elif conversion_type == "pennylane":
        from mitiq.interface.mitiq_pennylane.conversions import to_pennylane

        conversion_function = to_pennylane
    elif conversion_type == "qibo":
        from mitiq.interface.mitiq_qibo.conversions import to_qibo

        conversion_function = to_qibo
    elif conversion_type in FROM_MITIQ_DICT:
        conversion_function = FROM_MITIQ_DICT[conversion_type]

    elif conversion_type == "cirq":

        def conversion_function(circ: cirq.Circuit) -> cirq.Circuit:
            return circ

    else:
        raise UnsupportedCircuitError(
            f"Conversion to circuit type {conversion_type} is unsupported."
            f"\n\n Please register converters with"
            f"register_mitiq_converters(),"
            f"\n or specify a supported Circuit type:"
            f"\n {SUPPORTED_PROGRAM_TYPES}"
        )

    try:
        converted_circuit = conversion_function(circuit)
    except Exception:
        raise CircuitConversionError(
            f"Circuit could not be converted from an internal Mitiq type to a "
            f"circuit of type {conversion_type}."
        )

    return converted_circuit


P = ParamSpec("P")
R = TypeVar("R")


def accept_any_qprogram_as_input(
    accept_cirq_circuit_function: Callable[Concatenate[cirq.Circuit, P], R],
) -> Callable[Concatenate[QPROGRAM, P], R]:
    """Converts functions which take as input cirq.Circuit object (and return
    anything), to function which can accept any QPROGRAM.
    """

    @wraps(accept_cirq_circuit_function)
    def accept_any_qprogram_function(
        circuit: QPROGRAM, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        cirq_circuit, _ = convert_to_mitiq(circuit)
        return accept_cirq_circuit_function(cirq_circuit, *args, **kwargs)

    return accept_any_qprogram_function


def atomic_converter(
    cirq_circuit_modifier: Callable[..., Any],
) -> Callable[..., Any]:
    """Decorator which allows for a function which inputs and returns a Cirq
    circuit to input and return any QPROGRAM.

    Args:
        cirq_circuit_modifier: Function which inputs a Cirq circuit and returns
            a (potentially modified) Cirq circuit.
    """

    @wraps(cirq_circuit_modifier)
    def qprogram_modifier(
        circuit: QPROGRAM, *args: Any, **kwargs: Any
    ) -> QPROGRAM:
        # Convert to Mitiq representation.
        mitiq_circuit, input_circuit_type = convert_to_mitiq(circuit)

        # Modify the Cirq circuit.
        scaled_circuit = cirq_circuit_modifier(mitiq_circuit, *args, **kwargs)

        if kwargs.get("return_mitiq") is True:
            return scaled_circuit

        # Base conversion back to input type.
        scaled_circuit = convert_from_mitiq(scaled_circuit, input_circuit_type)

        return scaled_circuit

    return qprogram_modifier


def atomic_one_to_many_converter(
    cirq_circuit_modifier: Callable[..., Collection[cirq.Circuit]],
) -> Callable[..., Collection[QPROGRAM]]:
    """Convert function which returns multiple cirq.Circuits into a function
    which returns multiple QPROGRAM instances.
    """

    @wraps(cirq_circuit_modifier)
    def qprogram_modifier(
        circuit: QPROGRAM, *args: Any, **kwargs: Any
    ) -> Collection[QPROGRAM]:
        mitiq_circuit, input_circuit_type = convert_to_mitiq(circuit)

        modified_circuits = cirq_circuit_modifier(
            mitiq_circuit, *args, **kwargs
        )

        if kwargs.get("return_mitiq") is True:
            return modified_circuits

        return [
            convert_from_mitiq(modified_circuit, input_circuit_type)
            for modified_circuit in modified_circuits
        ]

    return qprogram_modifier


def accept_qprogram_and_validate(
    cirq_circuit_modifier: Callable[..., Any],
    one_to_many: bool = False,
) -> Callable[..., Any]:
    """This decorator performs two functions:

        1. Transforms a function of signature (cirq.Circuit -> cirq.Circuit)
        to (QPROGRAM -> QPROGRAM).
        2. Validates the incoming QPROGRAM instance to ensure Mitiq's error
        mitigation techniques can be applied to it.

    Args:
        cirq_circuit_modifier: The function modifying a Cirq circuit.
        one_to_many: If True, ``cirq_circuit_modifier`` is expected to return
            a sequence of Cirq circuits instead of a single Cirq circuit.

    Returns:
        The transformed function which can take any QPROGRAM, and performs
        circuit-level validation.
    """

    @wraps(cirq_circuit_modifier)
    def new_function(circuit: QPROGRAM, *args: Any, **kwargs: Any) -> QPROGRAM:
        # Pre atomic conversion
        if "qiskit" in circuit.__module__:
            from qiskit.transpiler.passes import RemoveBarriers

            from mitiq.interface.mitiq_qiskit.conversions import (
                _add_identity_to_idle,
            )

            # Avoid mutating the input circuit
            circuit = circuit.copy()
            # Removing barriers is necessary to correctly identify idle qubits
            circuit = RemoveBarriers()(circuit)
            # Apply identity gates to idle qubits otherwise they get lost
            # when converting to Cirq. Eventually, identities will be removed.
            idle_qubits = _add_identity_to_idle(circuit)

        if one_to_many:
            out_circuits = atomic_one_to_many_converter(cirq_circuit_modifier)(
                circuit, *args, **kwargs
            )
        else:
            out_circuit = atomic_converter(cirq_circuit_modifier)(
                circuit, *args, **kwargs
            )
            out_circuits = [out_circuit]

        circuits_to_return = []
        for out_circuit in out_circuits:
            # Post atomic conversion
            # PyQuil: Restore declarations, measurements, and metadata.
            if "pyquil" in out_circuit.__module__:
                from pyquil import Program
                from pyquil.quilbase import Declare, Measurement

                circuit = cast(Program, circuit)

                # Grab all measurements from the input circuit.
                measurements = [
                    instr
                    for instr in circuit.instructions
                    if isinstance(instr, Measurement)
                ]

                # Remove memory declarations added from Cirq-pyQuil conversion.
                new_declarations = {
                    k: v
                    for k, v in out_circuit.declarations.items()
                    if k == "ro" or v.memory_type != "BIT"
                }
                new_declarations.update(circuit.declarations)

                # Delete all declarations and measurements.
                instructions = [
                    instr
                    for instr in out_circuit.instructions
                    if not (isinstance(instr, (Declare, Measurement)))
                ]

                # Add back original declarations and measurements.
                out_circuit = Program(
                    list(new_declarations.values())
                    + instructions
                    + measurements
                )

                # Set the number of shots to the input circuit.
                out_circuit.num_shots = circuit.num_shots

            # Qiskit: Keep the same register structure and measurement order.
            if "qiskit" in out_circuit.__module__:
                from mitiq.interface.mitiq_qiskit.conversions import (
                    _measurement_order,
                    _remove_identity_from_idle,
                    _transform_registers,
                )

                out_circuit.remove_final_measurements()
                out_circuit = _transform_registers(
                    out_circuit, new_qregs=circuit.qregs
                )
                _remove_identity_from_idle(out_circuit, idle_qubits)
                if circuit.cregs and not out_circuit.cregs:
                    out_circuit.add_register(*circuit.cregs)

                for q, c in _measurement_order(circuit):
                    out_circuit.measure(q, c)
            circuits_to_return.append(out_circuit)

        if not one_to_many:
            assert len(circuits_to_return) == 1
            return circuits_to_return[0]

        return circuits_to_return

    return new_function


@accept_qprogram_and_validate
def append_cirq_circuit_to_qprogram(
    circuit: QPROGRAM, cirq_circuit: cirq.Circuit
) -> QPROGRAM:
    """Appends a Cirq circuit to a QPROGRAM."""
    return circuit + cirq_circuit



# Copyright (C) Unitary Foundation
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Unit tests for conversions between Mitiq circuits and Qiskit circuits."""

import copy

import cirq
import numpy as np
import pytest
import qiskit
from qiskit import qasm2

from mitiq.interface import convert_to_mitiq
from mitiq.interface.mitiq_qiskit.conversions import (
    _add_identity_to_idle,
    _map_bit_index,
    _measurement_order,
    _remove_identity_from_idle,
    _remove_qasm_barriers,
    _transform_registers,
    from_qasm,
    from_qiskit,
    to_qasm,
    to_qiskit,
)
from mitiq.utils import _equal


def _multi_reg_circuits():
    """Returns a circuit with multiple registers
    and an equivalent circuit with a single register."""
    qregs = [qiskit.QuantumRegister(1, f"q{j}") for j in range(4)]
    circuit_multi_reg = qiskit.QuantumCircuit(*qregs)
    for q in qregs[1:]:
        circuit_multi_reg.x(q)
        circuit_multi_reg.x(3)
    qreg = qiskit.QuantumRegister(4, "q")
    circuit_single_reg = qiskit.QuantumCircuit(qreg)
    for q in qreg[1:]:
        circuit_single_reg.x(q)
        circuit_single_reg.x(3)

    return circuit_multi_reg, circuit_single_reg


def test_remove_qasm_barriers():
    assert (
        _remove_qasm_barriers(
            """
// quantum ripple-carry adder from Cuccaro et al, quant-ph/0410184
OPENQASM 2.0;
include "qelib1.inc";
include "barrier.inc";
include ";barrier.inc";
gate majority a,b,c
{
  cx c,b;
  cx c,a;
  ccx a,b,c;
}
gate barrier1 a,a,a,a,a{
    barrier x,y,z;
    barrier1;
    barrier;
}
gate unmaj a,b,c
{
  ccx a,b,c;
  cx c,a;
  cx a,b;
}
qreg cin[1];
qreg a[4];
qreg b[4];
// barrier;
qreg cout[1]; barrier x,y,z;
creg ans[5];
// set input states
x a[0]; // a = 0001
x b;    // b = 1111
// add a to b, storing result in b
majority cin[0],b[0],a[0];
majority a[0],b[1],a[1];
majority a[1],b[2],a[2];
majority a[2],b[3],a[3];
cx a[3],cout[0];
unmaj a[2],b[3],a[3];
unmaj a[1],b[2],a[2];
unmaj a[0],b[1],a[1];
unmaj cin[0],b[0],a[0];
measure b[0] -> ans[0];
measure b[1] -> ans[1];
measure b[2] -> ans[2];
measure b[3] -> ans[3];
measure cout[0] -> ans[4];
// quantum ripple-carry adder from Cuccaro et al, quant-ph/0410184
"""
        )
        == """
// quantum ripple-carry adder from Cuccaro et al, quant-ph/0410184
OPENQASM 2.0;
include "qelib1.inc";
include "barrier.inc";
include ";barrier.inc";
gate majority a,b,c
{
  cx c,b;
  cx c,a;
  ccx a,b,c;
}
gate barrier1 a,a,a,a,a{
    barrier1;
}
gate unmaj a,b,c
{
  ccx a,b,c;
  cx c,a;
  cx a,b;
}
qreg cin[1];
qreg a[4];
qreg b[4];
// barrier;
qreg cout[1];
creg ans[5];
// set input states
x a[0]; // a = 0001
x b;    // b = 1111
// add a to b, storing result in b
majority cin[0],b[0],a[0];
majority a[0],b[1],a[1];
majority a[1],b[2],a[2];
majority a[2],b[3],a[3];
cx a[3],cout[0];
unmaj a[2],b[3],a[3];
unmaj a[1],b[2],a[2];
unmaj a[0],b[1],a[1];
unmaj cin[0],b[0],a[0];
measure b[0] -> ans[0];
measure b[1] -> ans[1];
measure b[2] -> ans[2];
measure b[3] -> ans[3];
measure cout[0] -> ans[4];
// quantum ripple-carry adder from Cuccaro et al, quant-ph/0410184
"""
    )


def test_bell_state_to_from_circuits():
    """Tests cirq.Circuit --> qiskit.QuantumCircuit --> cirq.Circuit
    with a Bell state circuit.
    """
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit(
        [cirq.ops.H.on(qreg[0]), cirq.ops.CNOT.on(qreg[0], qreg[1])]
    )
    qiskit_circuit = to_qiskit(cirq_circuit)  # Qiskit from Cirq
    circuit_cirq = from_qiskit(qiskit_circuit)  # Cirq from Qiskit
    assert _equal(cirq_circuit, circuit_cirq)


def test_bell_state_to_from_qasm():
    """Tests cirq.Circuit --> QASM string --> cirq.Circuit
    with a Bell state circuit.
    """
    qreg = cirq.LineQubit.range(2)
    cirq_circuit = cirq.Circuit(
        [cirq.ops.H.on(qreg[0]), cirq.ops.CNOT.on(qreg[0], qreg[1])]
    )
    qasm = to_qasm(cirq_circuit)  # Qasm from Cirq
    circuit_cirq = from_qasm(qasm)
    assert _equal(cirq_circuit, circuit_cirq)


def test_random_circuit_to_from_circuits():
    """Tests cirq.Circuit --> qiskit.QuantumCircuit --> cirq.Circuit
    with a random two-qubit circuit.
    """
    cirq_circuit = cirq.testing.random_circuit(
        qubits=2, n_moments=10, op_density=0.99, random_state=1
    )
    qiskit_circuit = to_qiskit(cirq_circuit)
    circuit_cirq = from_qiskit(qiskit_circuit)
    assert cirq.equal_up_to_global_phase(
        cirq_circuit.unitary(), circuit_cirq.unitary()
    )


def test_random_circuit_to_from_qasm():
    """Tests cirq.Circuit --> QASM string --> cirq.Circuit
    with a random one-qubit circuit.
    """
    cirq_circuit = cirq.testing.random_circuit(
        qubits=2, n_moments=10, op_density=0.99, random_state=2
    )
    qasm = to_qasm(cirq_circuit)
    circuit_cirq = from_qasm(qasm)
    assert cirq.equal_up_to_global_phase(
        cirq_circuit.unitary(), circuit_cirq.unitary()
    )


def test_convert_with_qft():
    """Tests converting a Qiskit circuit with a QFT to a Cirq circuit."""
    circuit = qiskit.QuantumCircuit(1)
    circuit &= qiskit.circuit.library.QFT(1)
    circuit.measure_all()
    qft_cirq = from_qiskit(circuit)
    qreg = cirq.LineQubit.range(1)
    cirq_circuit = cirq.Circuit(
        [cirq.ops.H.on(qreg[0]), cirq.ops.measure(qreg[0], key="meas")]
    )
    assert _equal(cirq_circuit, qft_cirq)


@pytest.mark.parametrize("as_qasm", (True, False))
def test_convert_with_barrier(as_qasm):
    """Tests converting a Qiskit circuit with a barrier to a Cirq circuit."""
    n = 5
    qiskit_circuit = qiskit.QuantumCircuit(qiskit.QuantumRegister(n))
    qiskit_circuit.barrier()

    if as_qasm:
        cirq_circuit = from_qasm(qasm2.dumps(qiskit_circuit))
    else:
        cirq_circuit = from_qiskit(qiskit_circuit)

    assert _equal(cirq_circuit, cirq.Circuit())


@pytest.mark.parametrize("as_qasm", (True, False))
def test_convert_with_multiple_barriers(as_qasm):
    """Tests converting a Qiskit circuit with barriers to a Cirq circuit."""
    n = 1
    num_ops = 10

    qreg = qiskit.QuantumRegister(n)
    qiskit_circuit = qiskit.QuantumCircuit(qreg)
    for _ in range(num_ops):
        qiskit_circuit.h(qreg)
        qiskit_circuit.barrier()

    if as_qasm:
        cirq_circuit = from_qasm(qasm2.dumps(qiskit_circuit))
    else:
        cirq_circuit = from_qiskit(qiskit_circuit)

    qbit = cirq.LineQubit(0)
    correct = cirq.Circuit(cirq.ops.H.on(qbit) for _ in range(num_ops))
    assert _equal(cirq_circuit, correct)


@pytest.mark.parametrize("reg_sizes", [[2, 4, 1, 6], [5, 4, 2], [6]])
def test_map_bit_index(reg_sizes):
    expected_register_index = 0
    expected_mapped_index = 0
    for bit_index in range(sum(reg_sizes)):
        register_index, mapped_index = _map_bit_index(bit_index, reg_sizes)

        assert register_index == expected_register_index
        assert mapped_index == expected_mapped_index

        expected_mapped_index += 1
        if bit_index == sum(reg_sizes[: expected_register_index + 1]) - 1:
            expected_register_index += 1
            expected_mapped_index = 0


@pytest.mark.parametrize("nqubits", [1, 5])
@pytest.mark.parametrize("with_ops", [True, False])
@pytest.mark.parametrize("measure", [True, False])
def test_transform_qregs_one_qubit_ops(nqubits, with_ops, measure):
    qreg = qiskit.QuantumRegister(nqubits)
    circ = qiskit.QuantumCircuit(qreg)
    if with_ops:
        circ.h(qreg)
    if measure:
        circ.add_register(qiskit.ClassicalRegister(nqubits))
        circ.measure(qreg, circ.cregs[0])

    orig = circ.copy()
    assert circ.qregs == [qreg]

    new_qregs = [qiskit.QuantumRegister(1) for _ in range(nqubits)]
    circ = _transform_registers(circ, new_qregs=new_qregs)

    assert circ.qregs == new_qregs
    assert circ.cregs == orig.cregs
    assert _equal(from_qiskit(circ), from_qiskit(orig))


@pytest.mark.parametrize("nqubits", [1, 5])
@pytest.mark.parametrize("with_ops", [True, False])
@pytest.mark.parametrize("measure", [True, False])
def test_transform_circuit_with_multiple_qregs(nqubits, with_ops, measure):
    qreg_1 = qiskit.QuantumRegister(nqubits)
    qreg_2 = qiskit.QuantumRegister(nqubits)
    circ = qiskit.QuantumCircuit(qreg_1, qreg_2)
    if with_ops:
        circ.h(qreg_1)
        circ.h(qreg_2)
    if measure:
        circ.add_register(qiskit.ClassicalRegister(nqubits))
        circ.measure(qreg_1, circ.cregs[0])
        circ.measure(qreg_2, circ.cregs[0])

    orig = circ.copy()
    assert circ.qregs == [qreg_1, qreg_2]

    new_qregs_1 = [qiskit.QuantumRegister(1) for _ in range(nqubits)]
    new_qregs_2 = [qiskit.QuantumRegister(1) for _ in range(nqubits)]
    circ = _transform_registers(circ, new_qregs=new_qregs_1 + new_qregs_2)

    assert circ.qregs == new_qregs_1 + new_qregs_2
    assert circ.cregs == orig.cregs
    assert _equal(from_qiskit(circ), from_qiskit(orig))


@pytest.mark.parametrize("new_reg_sizes", [[1], [1, 2], [2, 1], [1, 1, 1]])
def test_transform_qregs_two_qubit_ops(new_reg_sizes):
    nqubits = sum(new_reg_sizes)
    circ = to_qiskit(
        cirq.testing.random_circuit(
            nqubits, n_moments=5, op_density=1, random_state=1
        )
    )
    orig = circ.copy()

    new_qregs = [qiskit.QuantumRegister(s) for s in new_reg_sizes]
    circ = _transform_registers(circ, new_qregs=new_qregs)

    assert circ.qregs == new_qregs
    assert circ.cregs == orig.cregs
    assert _equal(from_qiskit(circ), from_qiskit(orig))


@pytest.mark.parametrize("new_reg_sizes", [[1], [1, 2], [2, 1], [1, 1, 1]])
@pytest.mark.parametrize("measure", [True, False])
def test_transform_qregs_random_circuit(new_reg_sizes, measure):
    nbits = sum(new_reg_sizes)
    circ = to_qiskit(
        cirq.testing.random_circuit(
            nbits, n_moments=5, op_density=1, random_state=10
        )
    )
    creg = qiskit.ClassicalRegister(nbits)
    circ.add_register(creg)
    if measure:
        circ.measure(circ.qregs[0], creg)
    orig = circ.copy()

    new_qregs = [qiskit.QuantumRegister(s) for s in new_reg_sizes]
    circ = _transform_registers(circ, new_qregs=new_qregs)

    assert circ.qregs == new_qregs
    assert _equal(from_qiskit(circ), from_qiskit(orig))


def test_transform_qregs_no_new_qregs():
    qreg = qiskit.QuantumRegister(5)
    circ = qiskit.QuantumCircuit(qreg)
    circ = _transform_registers(circ, new_qregs=None)
    assert circ.qregs == [qreg]


def test_transform_registers_too_few_qubits():
    circ = qiskit.QuantumCircuit(qiskit.QuantumRegister(2))
    new_qregs = [qiskit.QuantumRegister(1)]

    with pytest.raises(ValueError):
        _transform_registers(circ, new_qregs=new_qregs)


def test_transform_registers_adds_idle_qubits():
    """Tests transforming registers in a circuit with n qubits to a circuit
    with m > n qubits.
    """
    qreg = qiskit.QuantumRegister(1)
    creg = qiskit.ClassicalRegister(1)
    circuit = qiskit.QuantumCircuit(qreg, creg)
    circuit.x(qreg[0])
    circuit.measure(qreg[0], creg[0])

    assert len(circuit.qregs) == 1
    assert circuit.num_qubits == 1
    old_data = copy.deepcopy(circuit.data)

    circuit = _transform_registers(
        circuit, new_qregs=[qreg, qiskit.QuantumRegister(4)]
    )

    assert len(circuit.qregs) == 2
    assert circuit.num_qubits == 5
    assert circuit.data == old_data


def test_transform_registers_wrong_reg_number():
    nqubits = 2
    circ = qiskit.QuantumCircuit(qiskit.QuantumRegister(nqubits))
    new_qregs = [qiskit.QuantumRegister(1) for _ in range(2 * nqubits)]
    circ.add_register(*new_qregs)

    with pytest.raises(ValueError):
        _transform_registers(circ, new_qregs=new_qregs)


@pytest.mark.parametrize("size", [5])
def test_measurement_order(size):
    q, c = qiskit.QuantumRegister(size), qiskit.ClassicalRegister(size)
    circuit = qiskit.QuantumCircuit(q, c)

    index_order = [int(i) for i in np.random.RandomState(1).permutation(size)]
    for i in index_order:
        circuit.measure(q[i], c[i])

    order = _measurement_order(circuit)
    assert order == [(q[i], c[i]) for i in index_order]


def test_add_identity_to_idle():
    circuit = qiskit.QuantumCircuit(9)
    circuit.x([0, 8])
    circuit.cx(0, 8)
    expected_idle_qubits = circuit.qubits[1:-1]

    idle_qubits = _add_identity_to_idle(circuit)
    id_qubits = []
    for gates, qubits, cargs in circuit.get_instructions("id"):
        for qubit in qubits:
            id_qubits.append(qubit)
    assert idle_qubits == set(expected_idle_qubits)
    assert set(id_qubits) == idle_qubits


def test_remove_identity_from_idle():
    idle_indices = set(range(1, 8))
    circuit = qiskit.QuantumCircuit(9)
    circuit.x([0, 8])
    circuit.cx(0, 8)
    _remove_identity_from_idle(circuit, idle_indices)
    id_indices = []
    for gates, qubits, cargs in circuit.get_instructions("id"):
        for qubit in qubits:
            id_indices.append(qubit.index)
    assert id_indices == []


def test_add_identity_to_idle_with_multiple_registers():
    """Tests idle qubits are correctly detected even with many registers."""
    circuit_multi_reg, circuit_single_reg = _multi_reg_circuits()
    _add_identity_to_idle(circuit_multi_reg)
    _add_identity_to_idle(circuit_single_reg)

    # The result should be the same for both types of registers
    assert _equal(
        convert_to_mitiq(circuit_multi_reg)[0],
        convert_to_mitiq(circuit_single_reg)[0],
        require_qubit_equality=False,  # Qubit names can be different
    )


def test_remove_identity_from_idle_with_multiple_registers():
    """Tests identities are correctly removed even with many registers."""
    circuit_multi_reg, circuit_single_reg = _multi_reg_circuits()

    idle_qubits_multi = _add_identity_to_idle(circuit_multi_reg.copy())
    idle_qubits_single = _add_identity_to_idle(circuit_single_reg.copy())

    _remove_identity_from_idle(circuit_multi_reg, idle_qubits_multi)
    _remove_identity_from_idle(circuit_single_reg, idle_qubits_single)

    # Adding and removing identities should preserve the input
    input_multi, input_single = _multi_reg_circuits()
    assert circuit_multi_reg == input_multi
    assert circuit_single_reg == input_single


def test_convert_to_mitiq_with_rx_and_rzz():
    """Tests that convert_to_mitiq works with RX and RZZ gates."""
    test_qc = qiskit.QuantumCircuit(2)
    test_qc.rx(0.1, 0)
    test_qc.rzz(0.1, 0, 1)
    assert convert_to_mitiq(test_qc)


def test_convert_to_mitiq_with_rx_and_ryy():
    """
    Tests that convert_to_mitiq works with RX and RYY gates.
    """
    test_qc = qiskit.QuantumCircuit(2)
    test_qc.rx(0.1, 0)
    test_qc.ryy(0.1, 0, 1)
    assert convert_to_mitiq(test_qc)


def test_convert_to_mitiq_with_sx():
    """
    Tests that convert_to_mitiq works with SX gates.
    """
    test_qc = qiskit.QuantumCircuit(1)
    test_qc.sx(0)
    assert convert_to_mitiq(test_qc)


def test_convert_to_mitiq_with_u():
    """
    Tests that convert_to_mitiq works with U gates.
    """

    test_qc = qiskit.QuantumCircuit(1)
    test_qc.u(0.1, 0.2, 0.3, 0)
    assert convert_to_mitiq(test_qc)


def test_convert_to_mitiq_with_p():
    """
    Tests that convert_to_mitiq works with P gates.
    """
    circuit = qiskit.QuantumCircuit(1)
    circuit.p(np.pi / 4, 0)

    assert convert_to_mitiq(circuit)


def test_convert_to_mitiq_with_cu1():
    """
    Tests that convert_to_mitiq works with CU1 gates.
    """
    test_qc = qiskit.QuantumCircuit(2)
    test_qc.h(0)
    test_qc.h(1)
    cu1_gate = qiskit.circuit.library.CU1Gate(np.pi / 4)
    test_qc.append(cu1_gate, [0, 1])
    assert convert_to_mitiq(test_qc)


def test_convert_to_mitiq_with_ecrgate():
    """
    Tests that convert_to_mitiq works with ECR gates.
    """
    circuit = qiskit.QuantumCircuit(2)
    circuit.h(0)
    circuit.h(1)
    circuit.append(qiskit.circuit.library.ECRGate(), [0, 1])
    assert convert_to_mitiq(circuit)


def test_convert_to_mitiq_with_rxx_rzz_ecr():
    """
    Tests that convert_to_mitiq works with RXX, RZZ, and ECR gates.
    """
    circuit = qiskit.QuantumCircuit(2)
    circuit.sx(0)
    circuit.append(qiskit.circuit.library.RXXGate(np.pi / 3), [0, 1])
    circuit.append(qiskit.circuit.library.RZZGate(np.pi / 4), [0, 1])
    circuit.append(qiskit.circuit.library.ECRGate(), [0, 1])
    assert convert_to_mitiq(circuit)


def test_convert_to_mitiq_with_rzx_ryy_p():
    """
    Tests that convert_to_mitiq works with RZX, RYY, and P gates.
    """
    rotation_circuit = qiskit.QuantumCircuit(2)
    rotation_circuit.p(np.pi / 8, 0)
    rotation_circuit.append(qiskit.circuit.library.RZXGate(np.pi / 6), [0, 1])
    rotation_circuit.append(qiskit.circuit.library.RYYGate(np.pi / 5), [0, 1])
    assert convert_to_mitiq(rotation_circuit)


def test_convert_to_mitiq_with_qft_cu1_rzx():
    """
    Tests that convert_to_mitiq works with QFT, CU1, and RZX gates.
    """
    circuit = qiskit.QuantumCircuit(2)
    circuit.h(0)
    circuit.h(1)
    circuit.append(qiskit.circuit.library.QFT(2), [0, 1])
    circuit.append(qiskit.circuit.library.CU1Gate(np.pi / 3), [0, 1])
    circuit.append(qiskit.circuit.library.RZXGate(np.pi / 4), [0, 1])
    assert convert_to_mitiq(circuit)


def test_convert_to_mitiq_with_rzz_u_p_ecr():
    """
    Tests that convert_to_mitiq works with RZZ, U, P, and ECR gates.
    """
    circuit = qiskit.QuantumCircuit(2)
    circuit.append(qiskit.circuit.library.RZZGate(np.pi / 4), [0, 1])
    circuit.u(0.1, 0.2, 0.3, 0)
    circuit.p(np.pi / 4, 0)
    circuit.append(qiskit.circuit.library.ECRGate(), [0, 1])
    circuit.append(qiskit.circuit.library.RZZGate(np.pi / 2), [0, 1])
    assert convert_to_mitiq(circuit)


def test_convert_to_mitiq_with_rxx_ryy_sx_cu1():
    """
    Tests that convert_to_mitiq works with RXX, RYY, SX, and CU1 gates.
    """
    circuit = qiskit.QuantumCircuit(2)
    circuit.sx(0)
    circuit.append(qiskit.circuit.library.RXXGate(np.pi / 4), [0, 1])
    circuit.append(qiskit.circuit.library.RYYGate(np.pi / 6), [0, 1])
    circuit.append(qiskit.circuit.library.CU1Gate(np.pi / 8), [0, 1])
    circuit.u(0.5, 0.7, 0.2, 0)
    assert convert_to_mitiq(circuit)


def test_convert_to_mitiq_with_custom_operator():
    """
    Tests that convert_to_mitiq works with a custom operator.
    """
    gate = qiskit.quantum_info.Operator([[0.0, 1.0], [-1.0, 0.0]])
    qreg = qiskit.QuantumRegister(1)
    circ = qiskit.QuantumCircuit(qreg)
    circ.unitary(gate, [0])
    assert convert_to_mitiq(circ)
